
from .parallelbiflow import (
    BiFlowTrainCallable, BiFlowExecuteCallable,
    ParallelBiFlowException, ParallelBiFlow, ParallelCheckpointBiFlow)
from .parallelbihinet import ParallelCloneBiLayer

del parallelbiflow
del parallelbihinet



import os
import mdp

# wrap the mdp.test function and set the module path to bimdp path
infodict = mdp.NodeMetaclass._function_infodict(mdp.test)
idx = infodict["argnames"].index('mod_loc')
defaults = list(infodict['defaults'])
defaults[idx] = os.path.dirname(__file__)
infodict['defaults'] = tuple(defaults)

test = mdp.NodeMetaclass._wrap_function(mdp.test, infodict)



from builtins import range
from builtins import object
import mdp
from mdp import numx as n
from bimdp.nodes import SFABiNode, SFA2BiNode
from bimdp.parallel import ParallelBiFlow

# TODO: maybe test the helper classes as well, e.g. the new callable

class TestParallelBiNode(object):

    def test_stop_message_attribute(self):
        """Test that the stop_result attribute is present in forked node."""
        stop_result = ({"test": "blabla"}, "node123")
        x = n.random.random([100,10])
        node = SFABiNode(stop_result=stop_result)
        try:
            mdp.activate_extension("parallel")
            node2 = node.fork()
            node2.train(x)
            forked_result = node2.stop_training()
            assert forked_result == (None,) + stop_result
            # same with derived sfa2 node
            node = SFA2BiNode(stop_result=stop_result)
            mdp.activate_extension("parallel")
            node2 = node.fork()
            node2.train(x)
            forked_result = node2.stop_training()
            assert forked_result == (None,) + stop_result
        finally:
            mdp.deactivate_extension("parallel")


class TestParallelBiFlow(object):

    def test_nonparallel_flow(self):
        """Test a ParallelBiFlow with standard nodes."""
        flow = ParallelBiFlow([mdp.nodes.SFANode(output_dim=5),
                               mdp.nodes.PolynomialExpansionNode(degree=3),
                               mdp.nodes.SFANode(output_dim=20)])
        data_iterables = [[n.random.random((20,10)) for _ in range(6)],
                          None,
                          [n.random.random((20,10)) for _ in range(6)]]
        scheduler = mdp.parallel.Scheduler()
        flow.train(data_iterables, scheduler=scheduler)
        x = n.random.random([100,10])
        flow.execute(x)
        iterator = [n.random.random((20,10)) for _ in range(6)]
        flow.execute(iterator, scheduler=scheduler)
        scheduler.shutdown()

    def test_mixed_parallel_flow(self):
        """Test a ParallelBiFlow with both standard and BiNodes."""
        flow = ParallelBiFlow([mdp.nodes.PCANode(output_dim=8),
                               SFABiNode(output_dim=5),
                               SFA2BiNode(output_dim=20)])
        data_iterables = [[n.random.random((20,10)) for _ in range(6)]] * 3
        scheduler = mdp.parallel.Scheduler()
        flow.train(data_iterables, scheduler=scheduler)
        x = n.random.random([100,10])
        flow.execute(x)
        iterator = [n.random.random((20,10)) for _ in range(6)]
        flow.execute(iterator, scheduler=scheduler)
        scheduler.shutdown()

    def test_parallel_process(self):
        """Test training and execution with multiple training phases.

        The node with multiple training phases is a hinet.FlowNode.
        """
        sfa_node = mdp.nodes.SFANode(input_dim=10, output_dim=8)
        sfa2_node = mdp.nodes.SFA2Node(input_dim=8, output_dim=6)
        flow = ParallelBiFlow([sfa_node, sfa2_node])
        data_iterables = [[n.random.random((30,10)) for _ in range(6)],
                          [n.random.random((30,10)) for _ in range(7)]]
        scheduler = mdp.parallel.ProcessScheduler(n_processes=2)
        flow.train(data_iterables, scheduler=scheduler)
        flow.execute(data_iterables[1], scheduler=scheduler)
        x = n.random.random([100,10])
        flow.execute(x)
        iterator = [n.random.random((20,10)) for _ in range(6)]
        flow.execute(iterator, scheduler=scheduler)
        scheduler.shutdown()



"""
Package to inspect biflow training or execution by creating an HTML slideshow.
"""

from .tracer import (
    InspectionHTMLTracer, TraceHTMLConverter, TraceHTMLVisitor,
    TraceDebugException, inspection_css,
    prepare_training_inspection, remove_inspection_residues,
)
from .slideshow import (
    TrainHTMLSlideShow, SectExecuteHTMLSlideShow, ExecuteHTMLSlideShow
)
from .facade import (
    standard_css, EmptyTraceException,
    inspect_training, show_training, inspect_execution, show_execution
)

del tracer
del slideshow
del facade

from mdp.utils import fixup_namespace
fixup_namespace(__name__, None,
                ('tracer',
                 'slideshow',
                 'facade',
                 ))
del fixup_namespace




import mdp
from . import binode


class BiClassifier(binode.BiNode, mdp.ClassifierNode):
    """BiMDP version of the ClassifierNode base class.

    It enables that the classification results are returned by execute in the
    message.
    """

    def _execute(self, x, return_labels=None, return_probs=None,
                 return_ranks=None):
        """Return the unaltered x and classification results when requested.

        return_labels -- If True then the 'label' method is called on the x
            and the result is returned in the dict, with the key 'labels'. If
            return_labels is a string then this is used as a prefix for the
            'labels' key of the result.
        return_probs, return_ranks -- Work like return_labels, but the results
            are stored under the key 'probs' and 'ranks'.
        """
        msg = {}
        if return_labels:
            if not isinstance(return_labels, str):
                msg["labels"] = self.label(x)
            else:
                msg[return_labels + "labels"] = self.label(x)
        if return_probs:
            if not isinstance(return_probs, str):
                msg["probs"] = self.prob(x)
            else:
                msg[return_probs + "probs"] = self.prob(x)
        if return_ranks:
            if not isinstance(return_ranks, str):
                msg["ranks"] = self.rank(x)
            else:
                msg[return_ranks + "ranks"] = self.rank(x)
        if msg:
            return x, msg
        else:
            return x



"""
Special BiNode class derived from Node to allow complicated flow patterns.

Messages:
=========

The message argument 'msg' of the outer method 'execute' or 'train' is either a
dict or None (which is treated equivalently to an empty dict and is the default
value). The message is automatically parsed against the method signature of
_train or _execute (or any other specified method) in the following way:

    normal key string -- Is copied if in signature and passed as a named
        argument.

    node_id->key -- Is extracted (i.e. removed in original message) and passed
        as a named argument. The separator '->' is also stored available
        as the constant MSG_ID_SEP. If the key is not an argument of the
        message then the whole key is simply erased.

The msg returned from the inner part of the method (e.g. _execute) is then used
to update the original message (so values can be overwritten).

If args without default value are missing in the message, this will result in
the standard Python missing-arguments-exception (this is not checked by BiNode
itself).


BiNode Return Value Options:
============================

 result for execute:
     x, (x, msg), (x, msg, target)

 result for train:
    None -- terminates training
    x, (x, msg), (x, msg, target) -- Execution is continued and
        this node will be reached at a later time to terminate training.
        If the result has the form (None, msg) then the msg is dropped (so
        it is not required to 'clear' the message manually).

 result for stop_training:
    None -- Simply terminates the training, like for a normal node.
    x, (x, msg), (x, msg, target) -- Causes an execute like phase, which
        terminates when the end of the flow is reached or when EXIT_TARGET
        is given as target value (just like during a normal execute phase).


Magic message keys:
===================

When the incoming message is parsed by the BiNode base class, some argument
keywords are treated in a special way:

 'msg' -- If any method like _execute accept a 'msg' keyword then the complete
     remaining message (after parsing the other keywords) is supplied. The
     message in the return value then completely replaces the original message
     (instead of only updating it). This way a node can completely control the
     message and for example remove keys.

 'target' -- If any template method like execute finds a 'target' keyword in
     the message then this is used as the target value in the return value.
     However, if _execute then also returns a target value this overwrites the
     target value.

  'method' -- Specify the name of the method that should be used instead of
      the standard one (e.g. in execute the standard method is _execute).
      An underscore is automatically added in front, so to select _execute
      one would have to provide 'execute'.
      If 'inverse' is given then the inverse dimension check will be performed
      and if no target is provided it will be set to -1.

"""
from builtins import next
from builtins import str

import inspect

import mdp

# separator for node_id in message keys
MSG_ID_SEP = "->"


class BiNodeException(mdp.NodeException):
    """Exception for BiNode problems."""
    pass


class BiNode(mdp.Node):
    """Abstract base class for nodes that use bimdp features.

    This class itself is not non-functional.

    Derived class should, if necessary, overwrite the _bi_reset method
    (in addition to the normal mdp.Node methods). 

    Note hat this class can also be used as an Adapter / Mixin for normal nodes.
    This can for example be useful for nodes which require additional data
    arguments during training or execution. These can then be encapsulated in a
    messsage. Note that BiNode has to come first in the MRO to make all this
    work.
    """

    def __init__(self, node_id=None, stop_result=None, **kwargs):
        """Initialize BiNode.

        node_id -- None or string which identifies the node.
        stop_result -- A (msg, target) tupple which is used by stop_training.
            If _stop_training returns a result as well then is updates /
            overwrites the stop_result, otherwise simply stop_result is
            returned (with x set to None).
            If the node has multiple training phases then stop_result must be
            None or an iterable with one entry for each training phase.

        kwargs are forwarded via super to the next __init__ method
        in the MRO.
        """
        self._node_id = node_id
        self._stop_result = stop_result
        self._coroutine_instances = None
        super(BiNode, self).__init__(**kwargs)

    ### Modified template methods from mdp.Node. ###

    def execute(self, x, msg=None):
        """Return single value y or a result tuple.

        x can be None, then the usual checks are omitted.

        The possible return types are y, (y, msg), (y, msg, target)
        The outgoing msg carries forward the incoming message content.
        The last entry in a result tuple must not be None.
        y can be None if the result is a tuple.

        This template method normally calls the corresponding _execute method
        or another method as specified in the message (using the magic 'method'
        key.
        """
        if msg is None:
            if x is None:
                err = "Both x and msg are None."
                raise BiNodeException(err)
            return super(BiNode, self).execute(x)
        msg_id_keys = self._get_msg_id_keys(msg)
        target = self._extract_message_key("target", msg, msg_id_keys)
        method_name = self._extract_message_key("method", msg, msg_id_keys)
        method, target = self._get_method(method_name, self._execute, target)
        msg, arg_dict = self._extract_method_args(method, msg, msg_id_keys)
        # perform specific checks
        if x is not None:
            if (not method_name) or (method_name == "execute"):
                self._pre_execution_checks(x)
                x = self._refcast(x)
            # testing for the actual method allows nodes to delegate method
            # resolution to internal nodes by manipulating _get_method
            elif method == self._inverse:
                self._pre_inversion_checks(x)
        result = method(x, **arg_dict)
        return self._combine_result(result, msg, target)

    def train(self, x, msg=None):
        """Train and return None or more if the execution should continue.

        The possible return types are None, y, (y, msg), (y, msg, target).
        The last entry in a result tuple must not be None.
        y can be None if the result is a tuple.

        This template method normally calls the corresponding _train method
        or another method as specified in the message (using the magic 'method'
        key.

        Note that the remaining msg and taret values are only used if _train
        (or the requested method) returns something different from None
        (so an empty dict can be used to trigger continued execution).
        """
        # perform checks, adapted from Node.train
        if not self.is_trainable():
            raise mdp.IsNotTrainableException("This node is not trainable.")
        if not self.is_training():
            err = "The training phase has already finished."
            raise mdp.TrainingFinishedException(err)
        if msg is None:
            if x is None:
                err = "Both x and msg are None."
                raise BiNodeException(err)
            # no fall-back on Node.train because we might have a return value
            self._check_input(x)
            try:
                self._check_train_args(x)
            except TypeError:
                err = ("%s training seems to require " % str(self) +
                       "additional arguments, but none were given.")
                raise BiNodeException(err)
            self._train_phase_started = True
            x = self._refcast(x)
            return self._train_seq[self._train_phase][0](x)
        msg_id_keys = self._get_msg_id_keys(msg)
        target = self._extract_message_key("target", msg, msg_id_keys)
        method_name = self._extract_message_key("method", msg, msg_id_keys)
        default_method = self._train_seq[self._train_phase][0]
        method, target = self._get_method(method_name, default_method, target)
        msg, arg_dict = self._extract_method_args(method, msg, msg_id_keys)
        # perform specific checks
        if x is not None:
            if (not method_name) or (method_name == "train"):
                self._check_input(x)
                try:
                    self._check_train_args(x, **arg_dict)
                except TypeError:
                    err = ("The given additional arguments %s " %
                                str(list(arg_dict.keys())) +
                           "are not compatible with training %s." % str(self))
                    raise BiNodeException(err)
                self._train_phase_started = True
                x = self._refcast(x)
            elif method == self._inverse:
                self._pre_inversion_checks(x)
        result = method(x, **arg_dict)
        if result is None:
            return None
        result = self._combine_result(result, msg, target)
        if (isinstance(result, tuple) and len(result) == 2 and
            result[0] is None):
            # drop the remaining msg, so that no maual clearing is required
            return None
        return result

    def stop_training(self, msg=None):
        """Stop training phase and start an execute phase with a target.

        The possible return types are None, y, (y, msg), (y, msg, target).
        For None nothing more happens, the training phase ends like for a
        standard MDP node.
        If a return value is given then an excute phase is started.

        This template method normally calls a _stop_training method from
        self._train_seq.

        If a stop_result was given in __init__ then it is used but can be
        overwritten by the returned _stop_training result or by the 
        msg argument provided by the BiFlow.
        """
        # basic checks
        if self.is_training() and self._train_phase_started == False:
            raise mdp.TrainingException("The node has not been trained.")
        if not self.is_training():
            err = "The training phase has already finished."
            raise mdp.TrainingFinishedException(err)
        # call stop_training
        if not msg:
            result = self._train_seq[self._train_phase][1]()
            target = None
        else:
            msg_id_keys = self._get_msg_id_keys(msg)
            target = self._extract_message_key("target", msg, msg_id_keys)
            method_name = self._extract_message_key("method", msg, msg_id_keys)
            default_method = self._train_seq[self._train_phase][1]
            method, target = self._get_method(method_name,
                                              default_method, target)
            msg, arg_dict = self._extract_method_args(method, msg, msg_id_keys)
            result = method(**arg_dict)
        # close the current phase
        self._train_phase += 1
        self._train_phase_started = False
        # check if we have some training phase left
        if self.get_remaining_train_phase() == 0:
            self._training = False
        # use stored stop message and update it with the result
        if self._stop_result:
            if self.has_multiple_training_phases():
                stored_stop_result = self._stop_result[self._train_phase - 1]
            else:
                stored_stop_result = self._stop_result
            # make sure that the original dict in stored_stop_result is not
            # modified (this could have unexpected consequences in some cases)
            stored_msg = stored_stop_result[0].copy()
            if msg:
                stored_msg.update(msg)
            msg = stored_msg
            if target is None:
                target = stored_stop_result[1]
        return self._combine_result(result, msg, target)

    ## Additional new methods. ##

    @property
    def node_id(self):
        """Return the node id (should be string) or None."""
        return self._node_id

    def bi_reset(self):
        """Reset the node for the next data chunck.

        This template method calls the _bi_reset method.

        This method is automatically called by BiFlow after the processing of
        a data chunk is completed (during both training and execution).

        All temporary data should be deleted. The internal node structure can
        be reset for the next data chunk. This is especially important if this
        node is called multiple times for a single chunk and an internal state
        keeps track of the actions to be performed for each call.
        """
        if self._coroutine_instances is not None:
            # delete the instance attributes to unshadow the coroutine
            # initialization methods
            for key in self._coroutine_instances:
                delattr(self, key)
            self._coroutine_instances = None
        self._bi_reset()

    def _bi_reset(self):
        """Hook method, overwrite when needed."""
        pass

    def _request_node_id(self, node_id):
        """Return the node if it matches the provided node id.

        Otherwise the return value is None. In this default implementation
        self is returned if node_id == self._node_id.

        Use this method instead of directly accessing self._node_id.
        This allows a node to be associated with multiple node_ids. Otherwise
        node_ids would not work for container nodes like BiFlowNode.
        """
        if self._node_id == node_id:
            return self
        else:
            return None

    ### Helper methods for msg handling. ###

    def _get_msg_id_keys(self, msg):
        """Return the id specific message keys for this node.

        The format is [(key, fullkey),...].
        """
        msg_id_keys = []
        for fullkey in msg:
            if fullkey.find(MSG_ID_SEP) > 0:
                node_id, key = fullkey.split(MSG_ID_SEP)
                if node_id == self._node_id:
                    msg_id_keys.append((key, fullkey))
        return msg_id_keys

    @staticmethod
    def _extract_message_key(key, msg, msg_id_keys):
        """Extract and return the requested key from the message.

        Note that msg and msg_id_keys are modfied if the found key was
        node_id specific.
        """
        value = None
        if key in msg:
            value = msg[key]
        # check for node_id specific key and remove it from the msg
        for i, (_key, _fullkey) in enumerate(msg_id_keys):
            if key == _key:
                value = msg.pop(_fullkey)
                msg_id_keys.pop(i)
                break
        return value

    @staticmethod
    def _extract_method_args(method, msg, msg_id_keys):
        """Extract the method arguments form the message.

        Return the new message and a dict with the keyword arguments (the
        return of the message is done because it can be set to None).
        """
        arg_keys = inspect.getargspec(method)[0]
        arg_dict = dict((key, msg[key]) for key in msg if key in arg_keys)
        for key, fullkey in msg_id_keys:
            if key in arg_keys:
                arg_dict[key] = msg.pop(fullkey)
            else:
                del msg[fullkey]
        if "msg" in arg_keys:
            arg_dict["msg"] = msg
            msg = None
        return msg, arg_dict

    def _get_method(self, method_name, default_method, target=None):
        """Return the method to be called and the target return value.

        method_name -- as provided in msg (without underscore)
        default_method -- bound method object
        target -- return target value as provided in message or None

        If the chosen method is _inverse then the default target is -1.
        """
        if not method_name:
            method = default_method
        elif method_name == "inverse":
            method = self._inverse
            if target is None:
                target = -1
        else:
            method_name = "_" + method_name
            try:
                method = getattr(self, method_name)
            except AttributeError:
                err = ("The message requested a method named '%s', but "
                       "there is no such method." % method_name)
                raise BiNodeException(err)
        return method, target

    @staticmethod
    def _combine_result(result, msg, target):
        """Combine the execution result with the provided values.

        result -- x, (x, msg) or (x, msg, target)

        The values in result always has priority.
        """
        # overwrite result values if necessary and return
        if isinstance(result, tuple):
            if msg:
                if result[1]:
                    # combine outgoing msg and remaining msg values
                    msg.update(result[1])
                result = (result[0], msg) + result[2:]
            if (target is not None) and (len(result) == 2):
                # use given target if no target value was returned
                result += (target,)
            return result
        else:
            # result is only single array
            if (not msg) and (target is None):
                return result
            elif target is None:
                return result, msg
            else:
                return result, msg, target

    ### Overwrite Special Methods ###

    def __repr__(self):
        """BiNode version of the Node representation, adding the node_id."""
        name = type(self).__name__
        inp = "input_dim=%s" % str(self.input_dim)
        out = "output_dim=%s" % str(self.output_dim)
        if self.dtype is None:
            typ = 'dtype=None'
        else:
            typ = "dtype='%s'" % self.dtype.name
        node_id = self.node_id
        if node_id is None:
            nid = 'node_id=None'
        else:
            nid = 'node_id="%s"' % node_id
        args = ', '.join((inp, out, typ, nid))
        return name + '(' + args + ')'

    def __add__(self, other):
        """Adding binodes returns a BiFlow.

        If a normal Node or Flow is added to a BiNode then a BiFlow is
        returned.
        Note that if a flow is added then a deep copy is used (deep
        copies of the nodes are used).
        """
        # unfortunately the inline imports are required to avoid
        # a cyclic import (unless one adds a helper function somewhere else)
        if isinstance(other, mdp.Node):
            import bimdp
            return bimdp.BiFlow([self, other])
        elif isinstance(other, mdp.Flow):
            flow_copy = other.copy()
            import bimdp
            biflow = bimdp.BiFlow([self.copy()] + flow_copy.flow)
            return biflow
        else:
            # can delegate old cases
            return super(BiNode, self).__add__(other)


class PreserveDimBiNode(BiNode, mdp.PreserveDimNode):
    """BiNode version of the PreserveDimNode."""
    pass


### Helper Functions / Decorators ###

def binode_coroutine(args=None, defaults=()):
    """Decorator for the convenient definition of BiNode couroutines.
    
    This decorator takes care of all the boilerplate code to use a coroutine
    as a BiNode method for continuations (which is more elegant and convenient
    than using a a state machine implementation). 

    args -- List of string names of the additional arguments. Note that the
        standard 'x' array is always given as the first value. So if n args
        are requested the yield will return n+1 values.
    defaults -- Tuple of default values for the arguments. If this tuple has
        n elements, they correspond to the last n elements in 'args'
        (following the convention of inspect.getargspec).
    
    Internally there are three methods/functions:
        - The user defined function containing the original coroutine code.
          This is only stored in the decorator closure.
        - A new method ('_coroutine_initialization') with the name and
          signature  of the decorated coroutine, which internally handles the
          first initialization of the coroutine instance.
          This method is returned by the decorator.
        - A method with the signature specified by the 'args' for the
          decorator.
          After the coroutine has been initialized this
          method shadows the initialization method in the class instance
          (using an instance attribute to shadow the class attribute).
    """
    if args is None:
        args = ["self", "x"]
    else:
        args = ["self", "x"] + args
    def _binode_coroutine(coroutine):
        # the original coroutine is only stored in this closure
        infodict = mdp.NodeMetaclass._function_infodict(coroutine)
        original_name = infodict["name"]
        ## create the coroutine interface method
        def _coroutine_interface(self, *args):
            try:
                return self._coroutine_instances[original_name].send(args)
            except StopIteration as exception:
                delattr(self, original_name)
                del self._coroutine_instances[original_name]
                if len(exception.args):
                    return exception.args
                else:
                    return None
        # turn the signature into the one specified by the args
        interface_infodict = infodict.copy()
        interface_infodict["signature"] = ", ".join(args)
        interface_infodict["defaults"] = defaults
        coroutine_interface = mdp.NodeMetaclass._wrap_function(
                                    _coroutine_interface, interface_infodict)
        ## create the initialization method
        def _coroutine_initialization(self, *args):
            coroutine_instance = coroutine(self, *args)
            bound_coroutine_interface = coroutine_interface.__get__(
                                                        self, self.__class__)
            if self._coroutine_instances is None:
                self._coroutine_instances = dict()
            self._coroutine_instances[original_name] = coroutine_instance
            setattr(self, original_name, bound_coroutine_interface)
            try:
                return next(coroutine_instance)
            except StopIteration as exception:
                delattr(self, original_name)
                del self._coroutine_instances[original_name]
                if len(exception.args):
                    return exception.args
                else:
                    return None
        coroutine_initialization = mdp.NodeMetaclass._wrap_function(
                                    _coroutine_initialization, infodict)
        return coroutine_initialization
    return _binode_coroutine



from builtins import range
from ._tools import *

import mdp.parallel as parallel
n = numx

# TODO: add test that the callable is forked exactly once before a call?

def test_scheduler():
    """Test scheduler with 6 tasks."""
    scheduler = parallel.Scheduler()
    for i in range(6):
        scheduler.add_task(i, lambda x: x**2)
    results = scheduler.get_results()
    scheduler.shutdown()
    # check result
    results = n.array(results)
    assert n.all(results == n.array([0,1,4,9,16,25]))
    
def test_scheduler_manager():
    """Test context manager interface for scheduler."""
    with parallel.Scheduler() as scheduler:
        for i in range(6):
            scheduler.add_task(i, lambda x: x**2)
        results = scheduler.get_results()
    assert n.all(results == n.array([0,1,4,9,16,25]))
    
def test_scheduler_manager_exception():
    """Test context manager interface for scheduler in case of an exception."""
    log = []
    class TestSchedulerException(Exception): pass
    class TestScheduler(parallel.Scheduler):
        def _shutdown(self):
            log.append("shutdown")
        def _process_task(self, data, task_callable, task_index):
            raise TestSchedulerException()
    try:
        with TestScheduler() as scheduler:
            for i in range(6):
                scheduler.add_task(i, lambda x: x**2)
            scheduler.get_results()
    except TestSchedulerException:
        pass
    assert log == ["shutdown"]

def test_cpu_count():
    """Test the cpu_count helper function."""
    n_cpus = parallel.cpu_count()
    assert isinstance(n_cpus, int)


def test_thread_scheduler_flow():
    """Test thread scheduler with real Nodes."""
    precision = 6
    node1 = mdp.nodes.PCANode(output_dim=20)
    node2 = mdp.nodes.PolynomialExpansionNode(degree=1)
    node3 = mdp.nodes.SFANode(output_dim=10)
    flow = mdp.parallel.ParallelFlow([node1, node2, node3])
    parallel_flow = mdp.parallel.ParallelFlow(flow.copy()[:])
    scheduler = parallel.ThreadScheduler(verbose=False,
                                         n_threads=3)
    input_dim = 30
    scales = n.linspace(1, 100, num=input_dim)
    scale_matrix = mdp.numx.diag(scales)
    train_iterables = [n.dot(mdp.numx_rand.random((5, 100, input_dim)),
                             scale_matrix)
                       for _ in range(3)]
    parallel_flow.train(train_iterables, scheduler=scheduler)
    x = mdp.numx.random.random((10, input_dim))
    # test that parallel execution works as well
    # note that we need more chungs then processes to test caching
    parallel_flow.execute([x for _ in range(8)], scheduler=scheduler)
    scheduler.shutdown()
    # compare to normal flow
    flow.train(train_iterables)
    assert parallel_flow[0].tlen == flow[0].tlen
    y1 = flow.execute(x)
    y2 = parallel_flow.execute(x)
    assert_array_almost_equal(abs(y1), abs(y2), precision)




from __future__ import division
from builtins import range
from past.utils import old_div
__docformat__ = "restructuredtext en"

import mdp
from mdp import numx
from mdp.utils import mult, matmult, invert_exp_funcs2
from mdp.nodes import GrowingNeuralGasNode

def nmonomials(degree, nvariables):
    """Return the number of monomials of a given degree in a given number
    of variables."""
    return int(mdp.utils.comb(nvariables+degree-1, degree))

def expanded_dim(degree, nvariables):
    """Return the size of a vector of dimension ``nvariables`` after
    a polynomial expansion of degree ``degree``."""
    return int(mdp.utils.comb(nvariables+degree, degree))-1

class _ExpansionNode(mdp.Node):

    def __init__(self, input_dim = None, dtype = None):
        super(_ExpansionNode, self).__init__(input_dim, None, dtype)

    def expanded_dim(self, dim):
        return dim

    @staticmethod
    def is_trainable():
        return False

    @staticmethod
    def is_invertible():
        return False

    def _set_input_dim(self, n):
        self._input_dim = n
        self._output_dim = self.expanded_dim(n)

    def _set_output_dim(self, n):
        msg = "Output dim cannot be set explicitly!"
        raise mdp.NodeException(msg)

class PolynomialExpansionNode(_ExpansionNode):
    """Perform expansion in a polynomial space."""

    def __init__(self, degree, input_dim = None, dtype = None):
        """
        Input arguments:
        degree -- degree of the polynomial space where the input is expanded
        """
        self._degree = int(degree)
        super(PolynomialExpansionNode, self).__init__(input_dim, dtype)

    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node."""
        return (mdp.utils.get_dtypes('AllFloat') +
                mdp.utils.get_dtypes('AllInteger'))

    def expanded_dim(self, dim):
        """Return the size of a vector of dimension 'dim' after
        a polynomial expansion of degree 'self._degree'."""
        return expanded_dim(self._degree, dim)

    def _execute(self, x):
        degree = self._degree
        dim = self.input_dim
        n = x.shape[1]

        # preallocate memory
        dexp = numx.zeros((self.output_dim, x.shape[0]), dtype=self.dtype)
        # copy monomials of degree 1
        dexp[0:n, :] = x.T

        k = n
        prec_end = 0
        next_lens = numx.ones((dim+1, ), dtype=numx.int64)
        next_lens[0] = 0
        for i in range(2, degree+1):
            prec_start = prec_end
            prec_end += nmonomials(i-1, dim)
            prec = dexp[prec_start:prec_end, :]

            lens = next_lens[:-1].cumsum(axis=0)
            next_lens = numx.zeros((dim+1, ), dtype=numx.int64)
            for j in range(dim):
                factor = prec[lens[j]:, :]
                len_ = factor.shape[0]
                dexp[k:k+len_, :] = x[:, j] * factor
                next_lens[j+1] = len_
                k = k+len_

        return dexp.T

class QuadraticExpansionNode(PolynomialExpansionNode):
    """Perform expansion in the space formed by all linear and quadratic
    monomials.
    ``QuadraticExpansionNode()`` is equivalent to a
    ``PolynomialExpansionNode(2)``"""

    def __init__(self, input_dim = None, dtype = None):
        super(QuadraticExpansionNode, self).__init__(2, input_dim = input_dim,
                                                     dtype = dtype)

class RBFExpansionNode(mdp.Node):
    """Expand input space with Gaussian Radial Basis Functions (RBFs).

    The input data is filtered through a set of unnormalized Gaussian
    filters, i.e.::

       y_j = exp(-0.5/s_j * ||x - c_j||^2)

    for isotropic RBFs, or more in general::

       y_j = exp(-0.5 * (x-c_j)^T S^-1 (x-c_j))

    for anisotropic RBFs.
    """

    def __init__(self, centers, sizes, dtype = None):
        """
        :Arguments:
          centers
            Centers of the RBFs. The dimensionality
            of the centers determines the input dimensionality;
            the number of centers determines the output
            dimensionalities
          sizes
            Radius of the RBFs.

            ``sizes`` is a list with one element for each RBF, either
            a scalar (the variance of the RBFs for isotropic RBFs)
            or a covariance matrix (for anisotropic RBFs).
            If ``sizes`` is not a list, the same variance/covariance
            is used for all RBFs.
        """
        super(RBFExpansionNode, self).__init__(None, None, dtype)
        self._init_RBF(centers, sizes)

    @staticmethod
    def is_trainable():
        return False

    @staticmethod
    def is_invertible():
        return False

    def _init_RBF(self, centers, sizes):
        # initialize the centers of the RBFs
        centers = numx.array(centers, self.dtype)

        # define input/output dim
        self.set_input_dim(centers.shape[1])
        self.set_output_dim(centers.shape[0])

        # multiply sizes if necessary
        sizes = numx.array(sizes, self.dtype)
        if sizes.ndim==0 or sizes.ndim==2:
            sizes = numx.array([sizes]*self._output_dim)
        else:
            # check number of sizes correct
            if sizes.shape[0] != self._output_dim:
                msg = "There must be as many RBF sizes as centers"
                raise mdp.NodeException(msg)

        if numx.isscalar(sizes[0]):
            # isotropic RBFs
            self._isotropic = True
        else:
            # anisotropic RBFs
            self._isotropic = False

            # check size
            if (sizes.shape[1] != self._input_dim or
                sizes.shape[2] != self._input_dim):
                msg = ("Dimensionality of size matrices should be the same " +
                       "as input dimensionality (%d != %d)"
                       % (sizes.shape[1], self._input_dim))
                raise mdp.NodeException(msg)

            # compute inverse covariance matrix
            for i in range(sizes.shape[0]):
                sizes[i,:,:] = mdp.utils.inv(sizes[i,:,:])

        self._centers = centers
        self._sizes = sizes

    def _execute(self, x):
        y = numx.zeros((x.shape[0], self._output_dim), dtype = self.dtype)
        c, s = self._centers, self._sizes
        for i in range(self._output_dim):
            dist = x - c[i,:]
            if self._isotropic:
                tmp = old_div((dist**2.).sum(axis=1), s[i])
            else:
                tmp = (dist*matmult(dist, s[i,:,:])).sum(axis=1)
            y[:,i] = numx.exp(-0.5*tmp)
        return y

class GrowingNeuralGasExpansionNode(GrowingNeuralGasNode):
    """
    Perform a trainable radial basis expansion, where the centers and
    sizes of the basis functions are learned through a growing neural
    gas.

      positions of RBFs
        position of the nodes of the neural gas

      sizes of the RBFs
        mean distance to the neighbouring nodes.

    Important: Adjust the maximum number of nodes to control the
    dimension of the expansion.

    More information on this expansion type can be found in:
    B. Fritzke.
    Growing cell structures-a self-organizing network for unsupervised
    and supervised learning. Neural Networks 7, p. 1441--1460 (1994).
    """

    def __init__(self, start_poss=None, eps_b=0.2, eps_n=0.006, max_age=50,
                 lambda_=100, alpha=0.5, d=0.995, max_nodes=100,
                 input_dim=None, dtype=None):
        """
        For a full list of input arguments please check the documentation
        of GrowingNeuralGasNode.

        max_nodes (default 100) : maximum number of nodes in the
                                  neural gas, therefore an upper bound
                                  to the output dimension of the
                                  expansion.
        """
        # __init__ is overwritten only to reset the default for
        # max_nodes. The default of the GrowingNeuralGasNode is
        # practically unlimited, possibly leading to very
        # high-dimensional expansions.
        super(GrowingNeuralGasExpansionNode, self).__init__(
            start_poss=start_poss, eps_b=eps_b, eps_n=eps_n, max_age=max_age,
            lambda_=lambda_, alpha=alpha, d=d, max_nodes=max_nodes,
            input_dim=input_dim, dtype=dtype)

    def _set_input_dim(self, n):
        # Needs to be overwritten because GrowingNeuralGasNode would
        # fix the output dim to n here.
        self._input_dim = n

    def _set_output_dim(self, n):
        msg = "Output dim cannot be set explicitly!"
        raise mdp.NodeException(msg)

    @staticmethod
    def is_trainable():
        return True

    @staticmethod
    def is_invertible():
        return False

    def _stop_training(self):
        super(GrowingNeuralGasExpansionNode, self)._stop_training()

        # set the output dimension to the number of nodes of the neural gas
        self._output_dim = self.get_nodes_position().shape[0]

        # use the nodes of the learned neural gas as centers for a radial
        # basis function expansion.
        centers = self.get_nodes_position()

        # use the mean distances to the neighbours as size of the RBF expansion
        sizes = []

        for i,node in enumerate(self.graph.nodes):

            # calculate the size of the current RBF
            pos = node.data.pos
            sizes.append(numx.array([((pos-neighbor.data.pos)**2).sum()
                                     for neighbor in node.neighbors() ]).mean())

        # initialize the radial basis function expansion with centers and sizes
        self.rbf_expansion = mdp.nodes.RBFExpansionNode(centers = centers,
                                                        sizes = sizes)

    def _execute(self,x):
        return self.rbf_expansion(x)


class GeneralExpansionNode(_ExpansionNode):
    """Expands the input samples by applying to them one or more functions provided.
    
    The functions to be applied are specified by a list [f_0, ..., f_k], where 
    f_i, for 0 <= i <= k, denotes a particular function. 
    The input data given to these functions is a two-dimensional array and
    the output is another two-dimensional array. The dimensionality of the output 
    should depend only on the dimensionality of the input.
    Given a two-dimensional input array x, the output of the node 
    is then [f_0(x), ..., f_k(x)], that is, the concatenation of each one of 
    the computed arrays f_i(x).

    This node has been designed to facilitate nonlinear, fixed but arbitrary  
    transformations of the data samples within MDP flows.
        
    **Example**::
   
        >>> import mdp
        >>> from mdp import numx
        
        >>> def identity(x): return x
        
        >>> def u3(x): return numx.absolute(x)**3 #A simple nonlinear transformation
        
        >>> def norm2(x): #Computes the norm of each sample returning an Nx1 array
        >>>     return ((x**2).sum(axis=1)**0.5).reshape((-1,1)) 
          
        >>> x = numx.array([[-2., 2.], [0.2, 0.3], [0.6, 1.2]])
        >>> gen = mdp.nodes.GeneralExpansionNode(funcs=[identity, u3, norm2])
        >>> print(gen.execute(x))
        >>> [[-2.          2.          8.          8.          2.82842712]
        >>>  [ 0.2         0.3         0.008       0.027       0.36055513]
        >>>  [ 0.6         1.2         0.216       1.728       1.34164079]]

    Original code contributed by Alberto Escalante.
    """
    def __init__(self, funcs, input_dim = None, dtype = None):
        """
        Short argument description:

          ``funcs``
               list of functions f_i that realize the expansion.
        """
        self.funcs = funcs
        super(GeneralExpansionNode, self).__init__(input_dim, dtype)

    def expanded_dim(self, n):
        """The expanded dim is computed by directly applying the expansion
        functions f_i to a zero input of dimension n.
        """
        return int(self.output_sizes(n).sum())
    
    def output_sizes(self, n):
        """Return the individual output sizes of each expansion function
        when the input has lenght n."""
        sizes = numx.zeros(len(self.funcs), dtype=numx.int64)
        x = numx.zeros((1,n))
        for i, func in enumerate(self.funcs):
            outx = func(x)
            sizes[i] = outx.shape[1]
        return sizes

    @staticmethod
    def is_trainable():
        return False

    @staticmethod
    def is_invertible():
        return False

    def pseudo_inverse(self, x, use_hint=None):
        """Calculate a pseudo inverse of the expansion using
        scipy.optimize.

        ``use_hint``
               when calculating a pseudo inverse of the expansion,
               the hint determines the starting point for the approximation.
               For details on this parameter see the function 
               ``invert_exp_funcs2`` in ``mdp.utils.routines.py``.

        This method requires scipy."""

        try:
            app_x_2, app_ex_x_2 = invert_exp_funcs2(x,
                                                    self.input_dim,
                                                    self.funcs,
                                                    use_hint=use_hint,
                                                    k=0.000001)
            return app_x_2.astype(self.dtype)
        except NotImplementedError as exc:
            raise mdp.MDPException(exc)

    def _execute(self, x):
        if self.input_dim is None:
            self.set_input_dim(x.shape[1])

        num_samples = x.shape[0]
        sizes = self.output_sizes(self.input_dim)

        out = numx.zeros((num_samples, self.output_dim), dtype=self.dtype)

        current_pos = 0
        for i, func in enumerate(self.funcs):
            out[:,current_pos:current_pos+sizes[i]] = func(x)
            current_pos += sizes[i]
        return out

### old weave inline code to perform a quadratic expansion
# weave C code executed in the function QuadraticExpansionNode.execute
## _EXPANSION_POL2_CCODE = """
##   // first of all, copy the linear part
##   for( int i=0; i<columns; i++ ) {
##     for( int l=0; l<rows; l++ ) {
##       dexp(l,i) = x(l,i);
##     }
##   }

##   // then, compute all monomials of second degree
##   int k=columns;
##   for( int i=0; i<columns; i++ ) {
##     for( int j=i; j<columns; j++ ) {
##       for( int l=0; l<rows; l++ ) {
##         dexp(l,k) = x(l,i)*x(l,j);
##       }
##       k++;
##     }
##   }
## """

# it was called like that:
##     def execute(self, x):
##         mdp.Node.execute(self, x)

##         rows = x.shape[0]
##         columns = self.input_dim
##         # dexp is going to contain the expanded signal
##         dexp = numx.zeros((rows, self.output_dim), dtype=self._dtype)

##         # execute the inline C code
##         weave.inline(_EXPANSION_POL2_CCODE,['rows','columns','dexp','x'],
##                  type_factories = weave.blitz_tools.blitz_type_factories,
##                  compiler='gcc',extra_compile_args=['-O3']);

##         return dexp



# -*- coding:utf-8 -*-
__docformat__ = "restructuredtext en"

from .pca_nodes import WhiteningNode, PCANode
from .sfa_nodes import SFANode, SFA2Node
from .ica_nodes import ICANode, CuBICANode, FastICANode, TDSEPNode
from .neural_gas_nodes import GrowingNeuralGasNode, NeuralGasNode
from .expansion_nodes import (QuadraticExpansionNode, PolynomialExpansionNode,
                             RBFExpansionNode, GrowingNeuralGasExpansionNode,
                             GeneralExpansionNode)
from .fda_nodes import FDANode
from .em_nodes import FANode
from .misc_nodes import (IdentityNode, HitParadeNode, TimeFramesNode,
                        TimeDelayNode, TimeDelaySlidingWindowNode,
                        EtaComputerNode, NoiseNode, NormalNoiseNode,
                        CutoffNode, HistogramNode, AdaptiveCutoffNode)
from .isfa_nodes import ISFANode
from .rbm_nodes import RBMNode, RBMWithLabelsNode
from .regression_nodes import LinearRegressionNode
from .classifier_nodes import (SignumClassifier, PerceptronClassifier,
                              SimpleMarkovClassifier,
                              DiscreteHopfieldClassifier,
                              KMeansClassifier, GaussianClassifier,
                              NearestMeanClassifier, KNNClassifier)
from .jade import JADENode
from .nipals import NIPALSNode
from .lle_nodes import LLENode, HLLENode
from .xsfa_nodes import XSFANode, NormalizeNode

# import internals for use in test_suites
from .misc_nodes import OneDimensionalHitParade as _OneDimensionalHitParade
from .expansion_nodes import expanded_dim as _expanded_dim

__all__ = ['PCANode', 'WhiteningNode', 'NIPALSNode', 'FastICANode',
           'CuBICANode', 'TDSEPNode', 'JADENode', 'SFANode', 'SFA2Node',
           'ISFANode', 'XSFANode', 'FDANode', 'FANode', 'RBMNode',
           'RBMWithLabelsNode', 'GrowingNeuralGasNode', 'LLENode', 'HLLENode',
           'LinearRegressionNode', 'QuadraticExpansionNode',
           'PolynomialExpansionNode', 'RBFExpansionNode','GeneralExpansionNode',
           'GrowingNeuralGasExpansionNode', 'NeuralGasNode', '_expanded_dim',
           'SignumClassifier',
           'PerceptronClassifier', 'SimpleMarkovClassifier',
           'DiscreteHopfieldClassifier', 'KMeansClassifier', 'NormalizeNode',
           'GaussianClassifier', 'NearestMeanClassifier', 'KNNClassifier',
           'EtaComputerNode', 'HitParadeNode', 'NoiseNode', 'NormalNoiseNode',
           'TimeFramesNode', 'TimeDelayNode', 'TimeDelaySlidingWindowNode',
           'CutoffNode', 'AdaptiveCutoffNode', 'HistogramNode',
           'IdentityNode', '_OneDimensionalHitParade']

# nodes with external dependencies
from mdp import config, numx_description, MDPException

if numx_description == 'scipy':
    from .convolution_nodes import Convolution2DNode
    __all__ += ['Convolution2DNode']

if config.has_shogun:
    from .shogun_svm_classifier import ShogunSVMClassifier
    __all__ += ['ShogunSVMClassifier']

if config.has_libsvm:
    from .libsvm_classifier import LibSVMClassifier
    __all__ += ['LibSVMClassifier']

if config.has_sklearn:
    from . import scikits_nodes
    for name in scikits_nodes.DICT_:
        if name.endswith('Node'):
            globals()[name] = scikits_nodes.DICT_[name]
            __all__.append(name)
        del name

from mdp import utils
utils.fixup_namespace(__name__, __all__ + ['ICANode'],
                      ('pca_nodes',
                       'sfa_nodes',
                       'ica_nodes',
                       'neural_gas_nodes',
                       'expansion_nodes',
                       'fda_nodes',
                       'em_nodes',
                       'misc_nodes',
                       'isfa_nodes',
                       'rbm_nodes',
                       'regression_nodes',
                       'classifier_nodes',
                       'jade',
                       'nipals',
                       'lle_nodes',
                       'xsfa_nodes',
                       'convolution_nodes',
                       'shogun_svm_classifier',
                       'svm_classifiers',
                       'libsvm_classifier',
                       'regression_nodes',
                       'classifier_nodes',
                       'utils',
                       'scikits_nodes',
                       'numx_description',
                       'config',
                       ))



"""
Script to be called on a remote machine for starting a pp network server.

This script calls pp_slave_wrapper in a new process and returns the pid.
The ssh connection stays open and can be used to kill the server process.

The python_executable and the paths are send via stdin.
The first sys.argv argument ist the nice value.
The other arguments and the paths are then used as arguments for the
wrapper script.
"""
from __future__ import print_function

import sys
import subprocess

def main():
    try:
        # receive sys_paths via stdin to be used in the wrapper
        python_executable = sys.stdin.readline()[:-1] # remove newline character
        sys_paths = []
        while True:
            sys_path = sys.stdin.readline()[:-1] # remove newline character
            if sys_path == "_done_":
                break
            sys_paths.append(sys_path)
        # assemble the command line for the wrapper by forwarding the arguments and
        cmd = ("nice %s %s pp_slave_wrapper.py" %
               (sys.argv[1], python_executable))
        for arg in sys.argv[2:]:
            cmd += " " + arg
        for sys_path in sys_paths:
            cmd += " " + sys_path
        # start the subprocess in which the slave process runs
        proc = subprocess.Popen(cmd, shell=True,
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        # print status message from slave process
        sys.stdout.write(proc.stdout.readline())
        sys.stdout.flush()
        # return the pid via stdout
        print(proc.pid)
        sys.stdout.flush()
    except Exception as e:
        print("Error while starting the server process.")
        print(e)
        print(-1)
        sys.stdout.flush()

if __name__ == "__main__":
    main()



from builtins import range
from builtins import object
import mdp
import warnings

# import numeric module (scipy, Numeric or numarray)
numx = mdp.numx

def _check_roundoff(t, dtype):
    """Check if t is so large that t+1 == t up to 2 precision digits"""
    # limit precision
    limit = 10.**(numx.finfo(dtype).precision-2)
    if int(t) >= limit:
        wr = ('You have summed %e entries in the covariance matrix.'
              '\nAs you are using dtype \'%s\', you are '
              'probably getting severe round off'
              '\nerrors. See CovarianceMatrix docstring for more'
              ' information.' % (t, dtype.name))
        warnings.warn(wr, mdp.MDPWarning)

class CovarianceMatrix(object):
    """This class stores an empirical covariance matrix that can be updated
    incrementally. A call to the 'fix' method returns the current state of
    the covariance matrix, the average and the number of observations, and
    resets the internal data.

    Note that the internal sum is a standard __add__ operation. We are not
    using any of the fancy sum algorithms to avoid round off errors when
    adding many numbers. If you want to contribute a CovarianceMatrix class
    that uses such algorithms we would be happy to include it in MDP.
    For a start see the Python recipe by Raymond Hettinger at
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/393090
    For a review about floating point arithmetic and its pitfalls see
    http://docs.oracle.com/cd/E19957-01/806-3568/ncg_goldberg.html
    """

    def __init__(self, dtype=None, bias=False):
        """If dtype is not defined, it will be inherited from the first
        data bunch received by 'update'.
        All the matrices in this class are set up with the given dtype and
        no upcast is possible.
        If bias is True, the covariance matrix is normalized by dividing
        by T instead of the usual T-1.
        """
        if dtype is None:
            self._dtype = None
        else:
            self._dtype = numx.dtype(dtype)
        self._input_dim = None  # will be set in _init_internals
        # covariance matrix, updated during the training phase
        self._cov_mtx = None
        # average, updated during the training phase
        self._avg = None
        # number of observation so far during the training phase
        self._tlen = 0

        self.bias = bias

    def _init_internals(self, x):
        """Init the internal structures.

        The reason this is not done in the constructor is that we want to be
        able to derive the input dimension and the dtype directly from the
        data this class receives.
        """
        # init dtype
        if self._dtype is None:
            self._dtype = x.dtype
        dim = x.shape[1]
        self._input_dim = dim
        type_ = self._dtype
        # init covariance matrix
        self._cov_mtx = numx.zeros((dim, dim), type_)
        # init average
        self._avg = numx.zeros(dim, type_)

    def update(self, x):
        """Update internal structures.

        Note that no consistency checks are performed on the data (this is
        typically done in the enclosing node).
        """
        if self._cov_mtx is None:
            self._init_internals(x)
        # cast input
        x = mdp.utils.refcast(x, self._dtype)
        # update the covariance matrix, the average and the number of
        # observations (try to do everything inplace)
        self._cov_mtx += mdp.utils.mult(x.T, x)
        self._avg += x.sum(axis=0)
        self._tlen += x.shape[0]

    def fix(self, center=True):
        """Returns a triple containing the covariance matrix, the average and
        the number of observations. The covariance matrix is then reset to
        a zero-state.

        If center is false, the returned matrix is the matrix of the second moments,
        i.e. the covariance matrix of the data without subtracting the mean."""
        # local variables
        type_ = self._dtype
        tlen = self._tlen
        _check_roundoff(tlen, type_)
        avg = self._avg
        cov_mtx = self._cov_mtx

        ##### fix the training variables
        # fix the covariance matrix (try to do everything inplace)
        if self.bias:
            cov_mtx /= tlen
        else:
            cov_mtx /= tlen - 1

        if center:
            avg_mtx = numx.outer(avg, avg)
            if self.bias:
                avg_mtx /= tlen*(tlen)
            else:
                avg_mtx /= tlen*(tlen - 1)
            cov_mtx -= avg_mtx

        # fix the average
        avg /= tlen

        ##### clean up
        # covariance matrix, updated during the training phase
        self._cov_mtx = None
        # average, updated during the training phase
        self._avg = None
        # number of observation so far during the training phase
        self._tlen = 0

        return cov_mtx, avg, tlen


class DelayCovarianceMatrix(object):
    """This class stores an empirical covariance matrix between the signal and
    time delayed signal that can be updated incrementally.

    Note that the internal sum is a standard __add__ operation. We are not
    using any of the fancy sum algorithms to avoid round off errors when
    adding many numbers. If you want to contribute a CovarianceMatrix class
    that uses such algorithms we would be happy to include it in MDP.
    For a start see the Python recipe by Raymond Hettinger at
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/393090
    For a review about floating point arithmetic and its pitfalls see
    http://docs.sun.com/source/806-3568/ncg_goldberg.html
    """

    def __init__(self, dt, dtype=None, bias=False):
        """dt is the time delay. If dt==0, DelayCovarianceMatrix equals
        CovarianceMatrix. If dtype is not defined, it will be inherited from
        the first data bunch received by 'update'.
        All the matrices in this class are set up with the given dtype and
        no upcast is possible.
        If bias is True, the covariance matrix is normalized by dividing
        by T instead of the usual T-1.
        """

        # time delay
        self._dt = int(dt)

        if dtype is None:
            self._dtype = None
        else:
            self._dtype = numx.dtype(dtype)

        # clean up variables to spare on space
        self._cov_mtx = None
        self._avg = None
        self._avg_dt = None
        self._tlen = 0

        self.bias = bias

    def _init_internals(self, x):
        """Inits some internals structures. The reason this is not done in
        the constructor is that we want to be able to derive the input
        dimension and the dtype directly from the data this class receives.
        """

        # init dtype
        if self._dtype is None:
            self._dtype = x.dtype
        dim = x.shape[1]
        self._input_dim = dim
        # init covariance matrix
        self._cov_mtx = numx.zeros((dim, dim), self._dtype)
        # init averages
        self._avg = numx.zeros(dim, self._dtype)
        self._avg_dt = numx.zeros(dim, self._dtype)

    def update(self, x):
        """Update internal structures."""
        if self._cov_mtx is None:
            self._init_internals(x)

        # cast input
        x = mdp.utils.refcast(x, self._dtype)

        dt = self._dt

        # the number of data points in each block should be at least dt+1
        tlen = x.shape[0]
        if tlen < (dt+1):
            err = 'Block length is %d, should be at least %d.' % (tlen, dt+1)
            raise mdp.MDPException(err)

        # update the covariance matrix, the average and the number of
        # observations (try to do everything inplace)
        self._cov_mtx += mdp.utils.mult(x[:tlen-dt, :].T, x[dt:tlen, :])
        totalsum = x.sum(axis=0)
        self._avg += totalsum - x[tlen-dt:, :].sum(axis=0)
        self._avg_dt += totalsum - x[:dt, :].sum(axis=0)
        self._tlen += tlen-dt

    def fix(self, A=None):
        """The collected data is adjusted to compute the covariance matrix of
        the signal x(1)...x(N-dt) and the delayed signal x(dt)...x(N),
        which is defined as <(x(t)-<x(t)>)*(x(t+dt)-<x(t+dt)>)> .
        The function returns a tuple containing the covariance matrix,
        the average <x(t)> over the first N-dt points, the average of the
        delayed signal <x(t+dt)> and the number of observations. The internal
        data is then reset to a zero-state.

        If A is defined, the covariance matrix is transformed by the linear
        transformation Ax . E.g. to whiten the data, A is the whitening matrix.
        """

        # local variables
        type_ = self._dtype
        tlen = self._tlen
        _check_roundoff(tlen, type_)
        avg = self._avg
        avg_dt = self._avg_dt
        cov_mtx = self._cov_mtx

        ##### fix the training variables
        # fix the covariance matrix (try to do everything inplace)
        avg_mtx = numx.outer(avg, avg_dt)
        avg_mtx /= tlen

        cov_mtx -= avg_mtx
        if self.bias:
            cov_mtx /= tlen
        else:
            cov_mtx /= tlen - 1

        if A is not None:
            cov_mtx = mdp.utils.mult(A, mdp.utils.mult(cov_mtx, A.T))

        # fix the average
        avg /= tlen
        avg_dt /= tlen

        ##### clean up variables to spare on space
        self._cov_mtx = None
        self._avg = None
        self._avg_dt = None
        self._tlen = 0

        return cov_mtx, avg, avg_dt, tlen


class MultipleCovarianceMatrices(object):
    """Container class for multiple covariance matrices to easily
    execute operations on all matrices at the same time.
    Note: all operations are done in place where possible."""
    def __init__(self, covs):
        """Insantiate with a sequence of covariance matrices."""
        # swap axes to get the different covmat on to the 3rd axis
        self.dtype = covs[0].dtype
        self.covs = (numx.array(covs, dtype=self.dtype)).transpose([1, 2, 0])
        self.ncovs = len(covs)

    def __getitem__(self, item):
        return self.covs[:, :, item]

    def symmetrize(self):
        """Symmetrize matrices: C -> (C+C^T)/2 ."""
        # symmetrize cov matrices
        covs = self.covs
        covs = 0.5*(covs+covs.transpose([1, 0, 2]))
        self.covs = covs

    def weight(self, weights):
        """Apply a weighting factor to matrices.
        Argument can be a sequence or a single value. In the latter case
        the same weight is applied to all matrices."""
        # apply a weighting vector to cov matrices
        err = ("len(weights)=%d does not match number "
               "of matrices (%d)" % (len(weights), self.ncovs))
        assert len(weights) == self.ncovs, err
        self.covs *= mdp.utils.refcast(weights, self.dtype)

    def rotate(self, angle, indices):
        """Rotate matrices by angle in the plane defined by indices [i,j]."""
        covs = self.covs
        [i, j] = indices
        cos_ = numx.cos(angle)
        sin_ = numx.sin(angle)
        # rotate columns
        # you need to copy the first column that is modified
        covs_i = covs[:, i, :] + 0
        covs_j = covs[:, j, :]
        covs[:, i, :] =  cos_*covs_i - sin_*covs_j
        covs[:, j, :] =  sin_*covs_i + cos_*covs_j
        # rotate rows
        # you need to copy the first row that is modified
        covs_i = covs[i, :, :] + 0
        covs_j = covs[j, :, :]
        covs[i, :, :] =  cos_*covs_i - sin_*covs_j
        covs[j, :, :] =  sin_*covs_i + cos_*covs_j
        self.covs = covs

    def permute(self, indices):
        """Swap two columns and two rows of all matrices, whose indices are
        specified as [i,j]."""
        covs = self.covs
        [i, j] = indices
        covs[i, :, :], covs[j, :, :] = covs[j, :, :], covs[i, :, :] + 0
        covs[:, i, :], covs[:, j, :] = covs[:, j, :], covs[:, i, :] + 0
        self.covs = covs

    def transform(self, trans_matrix):
        """Apply a linear transformation to all matrices, defined by the
        transformation matrix."""
        trans_matrix = mdp.utils.refcast(trans_matrix, self.dtype)
        for cov in range(self.ncovs):
            self.covs[:, :, cov] = mdp.utils.mult(
                mdp.utils.mult(trans_matrix.T, self.covs[:, :, cov]),
                trans_matrix)

    def copy(self):
        """Return a deep copy of the instance."""
        return MultipleCovarianceMatrices(self.covs.transpose([2, 0, 1]))


class CrossCovarianceMatrix(CovarianceMatrix):

    def _init_internals(self, x, y):
        if self._dtype is None:
            self._dtype = x.dtype
            if y.dtype != x.dtype:
                err = 'dtype mismatch: x (%s) != y (%s)'%(x.dtype,
                                                          y.dtype)
                raise mdp.MDPException(err)
        dim_x = x.shape[1]
        dim_y = y.shape[1]
        type_ = self._dtype
        self._cov_mtx = numx.zeros((dim_x, dim_y), type_)
        self._avgx = numx.zeros(dim_x, type_)
        self._avgy = numx.zeros(dim_y, type_)


    def update(self, x, y):
        # check internal dimensions consistency
        if x.shape[0] != y.shape[0]:
            err = '# samples mismatch: x (%d) != y (%d)'%(x.shape[0],
                                                          y.shape[0])
            raise mdp.MDPException(err)

        if self._cov_mtx is None:
            self._init_internals(x, y)

        # cast input
        x = mdp.utils.refcast(x, self._dtype)
        y = mdp.utils.refcast(y, self._dtype)

        self._cov_mtx += mdp.utils.mult(x.T, y)
        self._avgx += x.sum(axis=0)
        self._avgy += y.sum(axis=0)
        self._tlen += x.shape[0]

    def fix(self):
        type_ = self._dtype
        tlen = self._tlen
        _check_roundoff(tlen, type_)
        avgx = self._avgx
        avgy = self._avgy
        cov_mtx = self._cov_mtx

        ##### fix the training variables
        # fix the covariance matrix (try to do everything inplace)
        avg_mtx = numx.outer(avgx, avgy)

        if self.bias:
            avg_mtx /= tlen*(tlen)
            cov_mtx /= tlen
        else:
            avg_mtx /= tlen*(tlen - 1)
            cov_mtx /= tlen - 1
        cov_mtx -= avg_mtx
        # fix the average
        avgx /= tlen
        avgy /= tlen

        ##### clean up
        # covariance matrix, updated during the training phase
        self._cov_mtx = None
        # average, updated during the training phase
        self._avgx = None
        self._avgy = None
        # number of observation so far during the training phase
        self._tlen = 0

        return cov_mtx, avgx, avgy, tlen



"""
Adapters for the Parallel Python library (http://www.parallelpython.com).

The PPScheduler class uses an existing pp scheduler and is a simple adapter.

LocalPPScheduler includes the creation of a local pp scheduler.
NetworkPPScheduler includes the management of the remote slaves via SSH.
"""

from __future__ import print_function

import sys
import os

import time
import subprocess
import signal
import traceback
import tempfile

from . import scheduling
import pp
import mdp

TEMPDIR_PREFIX='pp4mdp-monkeypatch.'

def _monkeypatch_pp(container_dir):
    """Apply a hack for http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=620551.

    Importing numpy fails because the parent directory of the slave
    script (/usr/share/pyshared) is added to the begging of sys.path.
    This is a temporary fix until parallel python or the way it is
    packaged in debian is changed.

    This function monkey-patches the ppworker module and changes the
    path to the slave script. A temporary directory is created and the
    worker script is copied there.

    The temporary directory should be automatically removed when this
    module is destroyed.

    XXX: remove this when parallel python or the way it is packaged in debian is changed.
    """
    import os.path, shutil

    # this part copied from pp.py, should give the same result hopefully
    ppworker = os.path.join(os.path.dirname(os.path.abspath(pp.__file__)),
                            'ppworker.py')

    global _ppworker_dir
    _ppworker_dir = mdp.utils.TemporaryDirectory(prefix=TEMPDIR_PREFIX, dir=container_dir)
    ppworker3 = os.path.join(_ppworker_dir.name, 'ppworker.py')
    shutil.copy(ppworker, ppworker3)

    mdp._pp_worker_command = pp._Worker.command[:]
    try:
        pp._Worker.command[pp._Worker.command.index(ppworker)] = ppworker3
    except TypeError:
        # pp 1.6.0 compatibility
        pp._Worker.command = pp._Worker.command.replace(ppworker, ppworker3)

if hasattr(mdp.config, 'pp_monkeypatch_dirname'):
    _monkeypatch_pp(mdp.config.pp_monkeypatch_dirname)

class PPScheduler(scheduling.Scheduler):
    """Adaptor scheduler for the parallel python scheduler.

    This scheduler is a simple wrapper for a pp server. A pp server instance
    has to be provided.
    """

    def __init__(self, ppserver, max_queue_length=1,
                 result_container=None, verbose=False):
        """Initialize the scheduler.

        ppserver -- Parallel Python Server instance.
        max_queue_length -- How long the queue can get before add_task blocks.
        result_container -- ResultContainer used to store the results.
            ListResultContainer by default.
        verbose -- If True to get progress reports from the scheduler.
        """
        if result_container is None:
            result_container = scheduling.ListResultContainer()
        super(PPScheduler, self).__init__(result_container=result_container,
                                          verbose=verbose)
        self.ppserver = ppserver
        self.max_queue_length = max_queue_length

    def _process_task(self, data, task_callable, task_index):
        """Non-blocking processing of tasks.

        Depending on the scheduler state this function is non-blocking or
        blocking. One reason for blocking can be a full task-queue.
        """
        task = (data, task_callable.fork(), task_index)
        def execute_task(task):
            """Call the first args entry and return the return value."""
            data, task_callable, task_index = task
            task_callable.setup_environment()
            return task_callable(data), task_index
        while True:
            if len(self.ppserver._Server__queue) > self.max_queue_length:
                # release lock for other threads and wait
                self._lock.release()
                time.sleep(0.5)
                self._lock.acquire()
            else:
                # release lock to enable result storage
                self._lock.release()
                # the inner tuple is a trick to prevent introspection by pp
                # this forces pp to simply pickle the object
                self.ppserver.submit(execute_task, args=(task,),
                                     callback=self._pp_result_callback)
                break

    def _pp_result_callback(self, result):
        """Calback method for pp to unpack the result and the task id.

        This method then calls the normal _store_result method.
        """
        if result is None:
            result = (None, None)
        self._store_result(*result)

    def _shutdown(self):
        """Call destroy on the ppserver."""
        self.ppserver.destroy()


class LocalPPScheduler(PPScheduler):
    """Uses a local pp server to distribute the work across cpu cores.

    The pp server is created automatically instead of being provided by the
    user (in contrast to PPScheduler).
    """

    def __init__(self, ncpus="autodetect", max_queue_length=1,
                 result_container=None, verbose=False):
        """Create an internal pp server and initialize the scheduler.

        ncpus -- Integer or 'autodetect', specifies the number of processes
            used.
        max_queue_length -- How long the queue can get before add_task blocks.
        result_container -- ResultContainer used to store the results.
            ListResultContainer by default.
        verbose -- If True to get progress reports from the scheduler.
        """
        ppserver = pp.Server(ncpus=ncpus)
        super(LocalPPScheduler, self).__init__(ppserver=ppserver,
                                          max_queue_length=max_queue_length,
                                          result_container=result_container,
                                          verbose=verbose)


# default secret
SECRET = "rosebud"

class NetworkPPScheduler(PPScheduler):
    """Scheduler which can manage pp remote servers (requires SSH).

    The remote slave servers are automatically started and killed at the end.

    Since the slaves are started via SSH this schduler does not work on normal
    Windows systems. On such systems you can start the pp slaves
    manually and then use the standard PPScheduler.
    """

    def __init__(self, max_queue_length=1,
                 result_container=None,
                 verbose=False,
                 remote_slaves=None,
                 source_paths=None,
                 port=50017,
                 secret=SECRET,
                 nice=-19,
                 timeout=3600,
                 n_local_workers=0,
                 slave_kill_filename=None,
                 remote_python_executable=None):
        """Initialize the remote slaves and create the internal pp scheduler.

        result_container -- ResultContainer used to store the results.
            ListResultContainer by default.
        verbose -- If True to get progress reports from the scheduler.
        remote_slaves -- List of tuples, the first tuple entry is a string
            containing the name or IP adress of the slave, the second entry
            contains the number of processes (i.e. the pp ncpus parameter).
            The second entry can be None to use 'autodetect'.
        source_paths -- List of paths that will be appended to sys.path in the
        slaves.
        n_local_workers -- Value of ncpus for this machine.
        secret -- Secret password to secure the remote slaves.
        slave_kill_filename -- Filename (including path) where a list of the
            remote slave processes should be stored. Together with the
            'kill_slaves' function this makes it possible to quickly all
            remote slave processes in case something goes wrong.
            If None, a tempfile is created.
        """
        self._remote_slaves = remote_slaves
        self._running_remote_slaves = None  # list of strings 'address:port'
        # list with processes for the ssh connections to the slaves
        self._ssh_procs = None
        self._remote_pids = None  # list of the pids of the remote servers
        self._port = port
        if slave_kill_filename is None:
            slave_kill_file = tempfile.mkstemp(prefix='MDPtmp-')[1]
        self.slave_kill_file = slave_kill_file
        self._secret = secret
        self._slave_nice = nice
        self._timeout = timeout
        if not source_paths:
            self._source_paths = []
        else:
            self._source_paths = source_paths
        if remote_python_executable is None:
            remote_python_executable = sys.executable
        self._python_executable = remote_python_executable
        module_file = os.path.abspath(__file__)
        self._script_path = os.path.dirname(module_file)
        self.verbose = verbose
        # start ppserver
        self._start_slaves()
        ppslaves = tuple(["%s:%d" % (address, self._port)
                          for address in self._running_remote_slaves])
        ppserver = pp.Server(ppservers=ppslaves,
                             ncpus=n_local_workers,
                             secret=self._secret)
        super(NetworkPPScheduler, self).__init__(ppserver=ppserver,
                                          max_queue_length=max_queue_length,
                                          result_container=result_container,
                                          verbose=verbose)

    def _shutdown(self):
        """Shutdown all slaves."""
        for ssh_proc in self._ssh_procs:
            os.kill(ssh_proc.pid, signal.SIGQUIT)
        super(NetworkPPScheduler, self)._shutdown()
        if self.verbose:
            print("All slaves shut down.")

    def start_slave(self, address, ncpus="autodetect"):
        """Start a single remote slave.

        The return value is a tuple of the ssh process handle and
        the remote pid.
        """
        try:
            print("starting slave " + address + " ...")
            proc = subprocess.Popen(["ssh","-T", "%s" % address],
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT)
            proc.stdin.write("cd %s\n" % self._script_path)
            cmd = (self._python_executable +
                   " pp_slave_script.py  %d %d %d %s %d" %
                   (self._slave_nice, self._port, self._timeout, self._secret,
                    ncpus))
            proc.stdin.write(cmd + "\n")
            # send additional information to the remote process
            proc.stdin.write(self._python_executable + "\n")
            for sys_path in self._source_paths:
                proc.stdin.write(sys_path + "\n")
            proc.stdin.write("_done_" + "\n")
            # print status message from slave
            sys.stdout.write(address + ": " + proc.stdout.readline())
            # get PID for remote slave process
            pid = None
            if self.verbose:
                print("*** output from slave %s ***" % address)
            while pid is None:
                # the slave process might first output some hello message
                try:
                    value = proc.stdout.readline()
                    if self.verbose:
                        print(value)
                    pid = int(value)
                except ValueError:
                    pass
            if self.verbose:
                print("*** output end ***")
            return (proc, pid)
        except:
            print("Initialization of slave %s has failed." % address)
            traceback.print_exc()
            return None

    def _start_slaves(self):
        """Start remote slaves.

        The slaves that could be started are stored in a textfile, in the form
        name:port:pid
        """
        with open(self.slave_kill_file, 'w') as slave_kill_file:
            self._running_remote_slaves = []
            self._remote_pids = []
            self._ssh_procs = []
            for (address, ncpus) in self._remote_slaves:
                ssh_proc, pid = self.start_slave(address, ncpus=ncpus)
                if pid is not None:
                    slave_kill_file.write("%s:%d:%d\n" %
                                          (address, pid, ssh_proc.pid))
                self._running_remote_slaves.append(address)
                self._remote_pids.append(pid)
                self._ssh_procs.append(ssh_proc)

def kill_slaves(slave_kill_filename):
    """Kill all remote slaves which are stored in the given file.

    This functions is only meant for emergency situations, when something
    went wrong and the slaves have to be killed manually.
    """
    with open(slave_kill_filename) as tempfile:
        for line in tempfile:
            address, pid, ssh_pid = line.split(":")
            pid = int(pid)
            ssh_pid = int(ssh_pid)
            # open ssh connection to to kill remote slave
            proc = subprocess.Popen(["ssh","-T", address],
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT)
            proc.stdin.write("kill %d\n" % pid)
            proc.stdin.flush()
            # kill old ssh connection
            try:
                os.kill(ssh_pid, signal.SIGKILL)
            except:
                pass
            # a kill might prevent the kill command transmission
            # os.kill(proc.pid, signal.SIGQUIT)
            print("killed slave " + address + " (pid %d)" % pid)
        print("all slaves killed.")


if __name__ == "__main__":
    if len(sys.argv) == 2:
        kill_slaves(sys.argv[1])
    else:
        sys.stderr.write("usage: %s slave_list.txt\n" % __file__)




from .autogen import binodes_code, biclassifiers_code
exec(binodes_code())
exec(biclassifiers_code())

from .miscnodes import IdentityBiNode, SenderBiNode
from .gradient import NotDifferentiableException, GradientExtensionNode



from __future__ import division
from builtins import range
from past.utils import old_div
__docformat__ = "restructuredtext en"

from mdp import numx, NodeException, Cumulator
from mdp.utils import mult
from mdp.nodes import PCANode
sqrt = numx.sqrt

class NIPALSNode(Cumulator, PCANode):
    """Perform Principal Component Analysis using the NIPALS algorithm.
    This algorithm is particularyl useful if you have more variable than
    observations, or in general when the number of variables is huge and
    calculating a full covariance matrix may be unfeasable. It's also more
    efficient of the standard PCANode if you expect the number of significant
    principal components to be a small. In this case setting output_dim to be
    a certain fraction of the total variance, say 90%, may be of some help.

    **Internal variables of interest**

      ``self.avg``
          Mean of the input data (available after training).

      ``self.d``
          Variance corresponding to the PCA components.

      ``self.v``
          Transposed of the projection matrix (available after training).
          
      ``self.explained_variance``
          When output_dim has been specified as a fraction of the total
          variance, this is the fraction of the total variance that is actually
          explained.

    Reference for NIPALS (Nonlinear Iterative Partial Least Squares):
    Wold, H.
    Nonlinear estimation by iterative least squares procedures.
    in David, F. (Editor), Research Papers in Statistics, Wiley,
    New York, pp 411-444 (1966).

    More information about Principal Component Analysis, a.k.a. discrete
    Karhunen-Loeve transform can be found among others in
    I.T. Jolliffe, Principal Component Analysis, Springer-Verlag (1986).

    Original code contributed by:
    Michael Schmuker, Susanne Lezius, and Farzad Farkhooi (2008).
    """
    def __init__(self, input_dim=None, output_dim=None, dtype=None,
                 conv = 1e-8, max_it = 100000):
        """
        The number of principal components to be kept can be specified as
        'output_dim' directly (e.g. 'output_dim=10' means 10 components
        are kept) or by the fraction of variance to be explained
        (e.g. 'output_dim=0.95' means that as many components as necessary
        will be kept in order to explain 95% of the input variance).

        Other Arguments:
           conv   - convergence threshold for the residual error.
           max_it - maximum number of iterations

        """
        super(NIPALSNode, self).__init__(input_dim, output_dim, dtype)
        self.conv = conv
        self.max_it = max_it

    def _train(self, x):
        super(NIPALSNode, self)._train(x)

    def _stop_training(self, debug=False):
        # debug argument is ignored but needed by the base class
        super(NIPALSNode, self)._stop_training()
        self._adjust_output_dim()
        if self.desired_variance is not None:
            des_var = True
        else:
            des_var = False

        X = self.data
        conv = self.conv
        dtype = self.dtype
        mean = X.mean(axis=0)
        self.avg = mean
        max_it = self.max_it
        tlen = self.tlen

        # remove mean
        X -= mean
        var = X.var(axis=0).sum()
        self.total_variance = var
        exp_var = 0

        eigenv = numx.zeros((self.input_dim, self.input_dim), dtype=dtype)
        d = numx.zeros((self.input_dim,), dtype = dtype)
        for i in range(self.input_dim):
            it = 0
            # first score vector t is initialized to first column in X
            t = X[:, 0]
            # initialize difference
            diff = conv + 1
            while diff > conv:
                # increase iteration counter
                it += 1
                # Project X onto t to find corresponding loading p
                # and normalize loading vector p to length 1
                p = old_div(mult(X.T, t),mult(t, t))
                p /= sqrt(mult(p, p))

                # project X onto p to find corresponding score vector t_new
                t_new = mult(X, p)
                # difference between new and old score vector
                tdiff = t_new - t
                diff = (tdiff*tdiff).sum()
                t = t_new
                if it > max_it:
                    msg = ('PC#%d: no convergence after'
                           ' %d iterations.'% (i, max_it))
                    raise NodeException(msg)

            # store ith eigenvector in result matrix
            eigenv[i, :] = p
            # remove the estimated principal component from X
            D = numx.outer(t, p)
            X -= D
            D = mult(D, p)
            d[i] = old_div((D*D).sum(),(tlen-1))
            exp_var += old_div(d[i],var)
            if des_var and (exp_var >= self.desired_variance):
                self.output_dim = i + 1
                break

        self.d = d[:self.output_dim]
        self.v = eigenv[:self.output_dim, :].T
        self.explained_variance = exp_var



from __future__ import division
from past.utils import old_div
from ._tools import *

def test_FANode():
    d = 10
    N = 5000
    k = 4

    mu = uniform((1, d))*3.+2.
    sigma = uniform((d,))*0.01
    #A = utils.random_rot(d)[:k,:]
    A = numx_rand.normal(size=(k,d))

    # latent variables
    y = numx_rand.normal(0., 1., size=(N, k))
    # observations
    noise = numx_rand.normal(0., 1., size=(N, d)) * sigma

    x = mult(y, A) + mu + noise

    fa = mdp.nodes.FANode(output_dim=k, dtype='d')
    fa.train(x)
    fa.stop_training()

    # compare estimates to real parameters
    assert_array_almost_equal(fa.mu[0,:], mean(x, axis=0), 5)
    assert_array_almost_equal(fa.sigma, std(noise, axis=0)**2, 2)
    # FA finds A only up to a rotation. here we verify that the
    # A and its estimation span the same subspace
    AA = numx.concatenate((A,fa.A.T),axis=0)
    u,s,vh = utils.svd(AA)
    assert sum(old_div(s,max(s))>1e-2)==k, \
           'A and its estimation do not span the same subspace'

    y = fa.execute(x)
    fa.generate_input()
    fa.generate_input(10)
    fa.generate_input(y)
    fa.generate_input(y, noise=True)

    # test that noise has the right mean and variance
    est = fa.generate_input(numx.zeros((N, k)), noise=True)
    est -= fa.mu
    assert_array_almost_equal(numx.diag(numx.cov(est, rowvar=0)),
                              fa.sigma, 3)
    assert_almost_equal(numx.amax(abs(numx.mean(est, axis=0)), axis=None), 0., 3)

    est = fa.generate_input(100000)
    assert_array_almost_equal_diff(numx.cov(est, rowvar=0),
                                   mdp.utils.mult(fa.A, fa.A.T), 1)

def test_FANode_indim():
    # FANode uses two slightly different initialization for input_dims
    # larger or smaller than 200
    x = numx_rand.normal(size=(5000, 10))
    mdp.nodes.FANode(output_dim=1)(x)
    x = numx_rand.normal(size=(5000, 500))
    mdp.nodes.FANode(output_dim=1)(x)

def test_FANode_singular_cov():
    x = numx.array([[ 1., 1., 0., 0., 0.],
                    [ 0., 1., 1., 0., 0.],
                    [ 0., 1., 0., 0., 0.],
                    [ 0., 1., 1., 0., 0.],
                    [ 0., 1., 0., 0., 1.],
                    [ 0., 1., 0., 1., 0.],
                    [ 0., 1., 0., 0., 0.],
                    [ 1., 1., 0., 0., 0.],
                    [ 1., 1., 0., 0., 0.],
                    [ 0., 1., 0., 1., 1.]])



    fanode = mdp.nodes.FANode(output_dim=3)
    fanode.train(x)
    # the matrix x is singular
    py.test.raises(mdp.NodeException, "fanode.stop_training()")




from __future__ import division
from builtins import range
from past.utils import old_div
import mdp
from ._tools import *

def test_RBM_sample_h():
    # number of visible and hidden units
    I, J = 2, 4

    # create RBM node
    bm = mdp.nodes.RBMNode(J, I)
    # fake training to initialize internals
    bm.train(numx.zeros((1,I)))
    # init to deterministic model
    bm.w[0,:] = [1,0,1,0]
    bm.w[1,:] = [0,1,0,1]
    bm.w *= 2e4
    bm.bv *= 0.
    bm.bh *= 0.

    # ### test 1
    v = numx.array([[0,0],[1,0],[0,1],[1,1.]])
    h = []
    for n in range(1000):
        prob, sample = bm.sample_h(v)
        h.append(sample)

    # check inferred probabilities
    expected_probs = numx.array([[0.5, 0.5, 0.5, 0.5],
                                  [1.0, 0.5, 1.0, 0.5],
                                  [0.5, 1.0, 0.5, 1.0],
                                  [1.0, 1.0, 1.0, 1.0]])
    assert_array_almost_equal(prob, expected_probs, 8)

    # check sampled units
    h = numx.array(h)
    for n in range(4):
        distr = h[:,n,:].mean(axis=0)
        assert_array_almost_equal(distr, expected_probs[n,:], 1)

    # ### test 2, with bias
    bm.bh -= 1e2
    h = []
    for n in range(100):
        prob, sample = bm.sample_h(v)
        h.append(sample)

    # check inferred probabilities
    expected_probs = numx.array([[0., 0., 0., 0.],
                                  [1.0, 0., 1.0, 0.],
                                  [0., 1.0, 0., 1.0],
                                  [1.0, 1.0, 1.0, 1.0]])
    assert_array_almost_equal(prob, expected_probs, 8)

    # check sampled units
    h = numx.array(h)
    for n in range(4):
        distr = h[:,n,:].mean(axis=0)
        assert_array_almost_equal(distr, expected_probs[n,:], 1)

def test_RBM_sample_v():
    # number of visible and hidden units
    I, J = 4, 2

    # create RBM node
    bm = mdp.nodes.RBMNode(J, I)
    # fake training to initialize internals
    bm.train(numx.zeros((1,I)))
    # init to deterministic model
    bm.w[:,0] = [1,0,1,0]
    bm.w[:,1] = [0,1,0,1]
    bm.w *= 2e4
    bm.bv *= 0
    bm.bh *= 0

    # test 1
    h = numx.array([[0,0],[1,0],[0,1],[1,1.]])
    v = []
    for n in range(1000):
        prob, sample = bm.sample_v(h)
        v.append(sample)

    # check inferred probabilities
    expected_probs = numx.array([[0.5, 0.5, 0.5, 0.5],
                                  [1.0, 0.5, 1.0, 0.5],
                                  [0.5, 1.0, 0.5, 1.0],
                                  [1.0, 1.0, 1.0, 1.0]])
    assert_array_almost_equal(prob, expected_probs, 8)

    # check sampled units
    v = numx.array(v)
    for n in range(4):
        distr = v[:,n,:].mean(axis=0)
        assert_array_almost_equal(distr, expected_probs[n,:], 1)

    # test 2, with bias
    bm.bv -= 1e2
    v = []
    for n in range(1000):
        prob, sample = bm.sample_v(h)
        v.append(sample)

    # check inferred probabilities
    expected_probs = numx.array([[0., 0., 0., 0.],
                                  [1.0, 0., 1.0, 0.],
                                  [0., 1.0, 0., 1.0],
                                  [1.0, 1.0, 1.0, 1.0]])
    assert_array_almost_equal(prob, expected_probs, 8)

    # check sampled units
    v = numx.array(v)
    for n in range(4):
        distr = v[:,n,:].mean(axis=0)
        assert_array_almost_equal(distr, expected_probs[n,:], 1)

def test_RBM_stability():
    # number of visible and hidden units
    I, J = 8, 2

    # create RBM node
    bm = mdp.nodes.RBMNode(J, I)
    bm._init_weights()
    # init to random model
    bm.w = mdp.utils.random_rot(max(I,J), dtype='d')[:I, :J]
    bm.bv = numx_rand.randn(I)
    bm.bh = numx_rand.randn(J)

    # save original weights
    real_w = bm.w.copy()
    real_bv = bm.bv.copy()
    real_bh = bm.bh.copy()

    # Gibbs sample to reach the equilibrium distribution
    N = int(1e4)
    v = numx_rand.randint(0,2,(N,I)).astype('d')
    for k in range(100):
        if k%5==0: spinner()
        p, h = bm._sample_h(v)
        p, v = bm._sample_v(h)

    # see that w remains stable after learning
    for k in range(100):
        if k%5==0: spinner()
        bm.train(v)
    bm.stop_training()

    assert_array_almost_equal(real_w, bm.w, 1)
    assert_array_almost_equal(real_bv, bm.bv, 1)
    assert_array_almost_equal(real_bh, bm.bh, 1)

def test_RBM_learning():
    # number of visible and hidden units
    I, J = 4, 2

    bm = mdp.nodes.RBMNode(J, I)
    bm.w = mdp.utils.random_rot(max(I,J), dtype='d')[:I, :J]

    # the observations consist of two disjunct patterns that
    # never appear together
    N = int(1e4)
    v = numx.zeros((N,I))
    for n in range(int(N)):
        r = numx_rand.random()
        if r>0.666: v[n,:] = [0,1,0,1]
        elif r>0.333: v[n,:] = [1,0,1,0]

    for k in range(1500):
        if k%5==0: spinner()

        if k>5:
            mom = 0.9
        else:
            mom = 0.5
        bm.train(v, epsilon=0.3, momentum=mom)
        if old_div(bm._train_err,N)<0.1: break
        #print '-------', bm._train_err

    assert old_div(bm._train_err, N) < 0.1

def _generate_data(bm, I, N):
    data = []
    h = numx.ones(I, dtype='d')
    for t in range(N):
        prob, v = bm._sample_v(h)
        prob, h = bm._sample_h(v)
        if (t > 500):
            data.append(v)

    return numx.asarray(data, dtype='d')

def test_RBM_bv_learning():
    # number of visible and hidden units
    I, J = 4, 4

    bm = mdp.nodes.RBMNode(J, I)
    bm._init_weights()
    # init to random biases, unit generation matrix
    bm.w = numx.eye(I, dtype='d')
    bm.bh *= 0.0
    bm.bv = numx.linspace(0.1, 0.9, I) * 5

    #### generate training data
    data = _generate_data(bm, I, 5000)

    #### learn from generated data
    train_bm = mdp.nodes.RBMNode(J, I)

    train_bm.train(data)
    train_bm.w = numx.eye(I, dtype='d')

    N = data.shape[0]
    for k in range(5000):
        if k%5==0: spinner()

        train_bm.train(data, epsilon=0.6, momentum=0.7)
        if abs(train_bm.bv - bm.bv).max() < 0.5: break

        # bv, bh, and w are dependent, so we need to keep one of them clamped
        train_bm.w = numx.eye(I, dtype='d')

    assert abs(train_bm.bv - bm.bv).max() < 0.5

def _test_RBM_bh_learning():
    # This one is tricky, as hidden biases are a very indirect parameter
    # of the input. We need to keep the rest of the weights clamped or there
    # would be alternative ways to explain the data

    # number of visible and hidden units
    I, J = 4, 4

    bm = mdp.nodes.RBMNode(J, I)
    bm._init_weights()
    # init to random biases, unit generation matrix
    bm.w = numx.eye(I, dtype='d')
    bm.bv *= 0.0
    bm.bh = numx.linspace(0.1, 0.9, I) * 5

    #### generate training data
    data = _generate_data(bm, I, 10000)

    #### learn from generated data
    train_bm = mdp.nodes.RBMNode(J, I)

    train_bm.train(data)
    train_bm.w = bm.w.copy()
    train_bm.bv *= 0.0

    N = data.shape[0]
    for k in range(5000):
        if k%5==0: spinner()

        train_bm.train(data, epsilon=3.0, momentum=0.8, update_with_ph=False)
        if abs(train_bm.bh - bm.bh).max() < 0.75: break

        # keep other weights clamped
        train_bm.w = bm.w.copy()
        train_bm.bv *= 0.0

    assert abs(train_bm.bh - bm.bh).max() < 0.75


def test_RBMWithLabelsNode():
    I, J, L = 4, 4, 2
    bm = mdp.nodes.RBMWithLabelsNode(J,L,I)
    assert bm.input_dim == I+L

    # generate input data
    N = 2500
    v = numx.zeros((2*N,I))
    l = numx.zeros((2*N,L))
    for n in range(N):
        r = numx_rand.random()
        if r>0.1:
            v[n,:] = [1,0,1,0]
            l[n,:] = [1,0]
    for n in range(N):
        r = numx_rand.random()
        if r>0.1:
            v[n,:] = [0,1,0,1]
            l[n,:] = [1,0]

    x = numx.concatenate((v, l), axis=1)
    for k in range(2500):
        if k%5==0: spinner()

        if k>200:
            mom = 0.9
            eps = 0.7
        else:
            mom = 0.5
            eps = 0.2
        bm.train(v, l, epsilon=eps, momentum=mom)

        ph, sh = bm._sample_h(x)
        pv, pl, sv, sl = bm._sample_v(sh, concatenate=False)

        v_train_err = float(((v-sv)**2.).sum())
        #print '-------', k, v_train_err/(2*N)
        if old_div(v_train_err, (2*N)) < 0.1:
            break

    # visible units are reconstructed
    assert old_div(v_train_err, (2*N)) < 0.1

    # units with 0 input have 50/50 labels
    idxzeros = v.sum(axis=1)==0
    nzeros = idxzeros.sum()
    point5 = numx.zeros((nzeros, L)) + 0.5
    assert_array_almost_equal(pl[idxzeros], point5, 2)



from __future__ import division
from builtins import range
from past.utils import old_div
from ._tools import *

def _std(x):
    return x.std(axis=0)
    # standard deviation without bias
    mx = mean(x, axis=0)
    mx2 = mean(x*x, axis=0)
    return numx.sqrt(old_div((mx2-mx),(x.shape[0]-1)))

def _cov(x,y=None):
    #return covariance matrix for x and y
    if y is None:
        y = x.copy()
    x = x - mean(x,0)
    x = old_div(x, _std(x))
    y = y - mean(y,0)
    y = old_div(y, _std(y))
    #return mult(numx.transpose(x),y)/(x.shape[0]-1)
    return old_div(mult(numx.transpose(x),y),(x.shape[0]))

def testISFANodeGivensRotations():
    ncovs = 5
    dim = 7
    ratio = uniform(2).tolist()
    covs = [uniform((dim,dim)) for j in range(ncovs)]
    covs= mdp.utils.MultipleCovarianceMatrices(covs)
    covs.symmetrize()
    i = mdp.nodes.ISFANode(list(range(1, ncovs+1)),sfa_ica_coeff=ratio,
                           icaweights=uniform(ncovs),
                           sfaweights=uniform(ncovs),
                           output_dim = dim-1, dtype="d")
    i._adjust_ica_sfa_coeff()
    ratio = i._bica_bsfa
    # case 2: only one axis within output space
    # get contrast using internal function
    phi, cont1, min_, dummy =\
               i._givens_angle_case2(dim-2,dim-1,covs,ratio,complete=1)
    # get contrast using explicit rotations
    cont2 = []
    for angle in phi:
        cp = covs.copy()
        cp.rotate(angle,[dim-2,dim-1])
        cont2.append(numx.sum(i._get_contrast(cp,ratio)))
    assert_array_almost_equal(cont1,cont2,decimal)
    # case 1: both axes within output space
    # get contrast using internal function
    phi,cont1, min_ , dummy =\
               i._givens_angle_case1(0,1,covs,ratio,complete = 1)
    # get contrast using explicit rotations
    cont2 = []
    for angle in phi:
        cp = covs.copy()
        cp.rotate(angle,[0,1])
        cont2.append(numx.sum(i._get_contrast(cp,ratio)))
    assert abs(min_) < old_div(numx.pi,4), 'Estimated Minimum out of bounds'
    assert_array_almost_equal(cont1,cont2,decimal)

def testISFANode_SFAPart():
    # create independent sources
    mat = uniform((100000,3))*2-1
    fmat = numx_fft.rfft(mat,axis=0)
    # enforce different speeds
    for i in range(3):
        fmat[(i+1)*5000:,i] = 0.
    mat = numx_fft.irfft(fmat,axis=0)
    _sfanode = mdp.nodes.SFANode()
    _sfanode.train(mat)
    src = _sfanode.execute(mat)
    # test with unmixed signals (i.e. the node should make nothing at all)
    out = mdp.nodes.ISFANode(lags=1,
                             whitened=True,
                             sfa_ica_coeff=[1.,0.])(src)
    max_cv = numx.diag(abs(_cov(out,src)))
    assert_array_almost_equal(max_cv, numx.ones((3,)),5)
    # mix linearly the signals
    mix = mult(src,uniform((3,3))*2-1)
    out = mdp.nodes.ISFANode(lags=1,
                             whitened=False,
                             sfa_ica_coeff=[1.,0.])(mix)
    max_cv = numx.diag(abs(_cov(out,src)))
    assert_array_almost_equal(max_cv, numx.ones((3,)),5)

def testISFANode_ICAPart():
    # create independent sources
    src = uniform((100000,3))*2-1
    fsrc = numx_fft.rfft(src,axis=0)
    # enforce different speeds
    for i in range(3):
        fsrc[(i+1)*5000:,i] = 0.
    src = numx_fft.irfft(fsrc,axis=0)
    # enforce time-lag-1-independence
    src = mdp.nodes.ISFANode(lags=1, sfa_ica_coeff=[1.,0.])(src)
    out = mdp.nodes.ISFANode(lags=1,
                             whitened=True,
                             sfa_ica_coeff=[0.,1.])(src)
    max_cv = numx.diag(abs(_cov(out,src)))
    assert_array_almost_equal(max_cv, numx.ones((3,)),5)
    # mix linearly the signals
    mix = mult(src,uniform((3,3))*2-1)
    out = mdp.nodes.ISFANode(lags=1,
                             whitened=False,
                             sfa_ica_coeff=[0.,1.])(mix)
    max_cv = numx.diag(abs(_cov(out,src)))
    assert_array_almost_equal(max_cv, numx.ones((3,)),5)

def testISFANode_3Complete():
    # test transition from ica to sfa behavior of isfa
    # use ad hoc sources
    lag = 25
    src = numx.zeros((1001,3),"d")
    idx = [(2,4),(80,1),(2+lag,6)]
    for i in range(len(idx)):
        i0, il = idx[i]
        src[i0:i0+il,i] = 1.
        src[i0+il:i0+2*il,i] = -1.
        src[:,i] -= mean(src[:,i])
        src[:,i] /= std(src[:,i])
    # test extreme cases
    # case 1: ICA
    out = mdp.nodes.ISFANode(lags=[1,lag],
                             icaweights=[1.,1.],
                             sfaweights=[1.,0.],
                             output_dim=2,
                             whitened=True,
                             sfa_ica_coeff=[1E-4,1.])(src)
    cv = abs(_cov(src,out))
    idx_cv = numx.argmax(cv,axis=0)
    assert_array_equal(idx_cv,[2,1])
    max_cv = numx.amax(cv,axis=0)
    assert_array_almost_equal(max_cv, numx.ones((2,)),5)
    # case 2: SFA
    out = mdp.nodes.ISFANode(lags=[1,lag],
                             icaweights=[1.,1.],
                             sfaweights=[1.,0.],
                             output_dim=2,
                             whitened=True,
                             sfa_ica_coeff=[1.,0.])(src)
    cv = abs(_cov(src,out))
    idx_cv = numx.argmax(cv,axis=0)
    assert_array_equal(idx_cv,[2,0])
    max_cv = numx.amax(cv,axis=0)
    assert_array_almost_equal(max_cv, numx.ones((2,)),5)

def _ISFA_analytical_solution( nsources, nmat, dim, ica_ambiguity):
    # build a sequence of random diagonal matrices
    matrices = [numx.eye(dim, dtype='d')]*nmat
    # build first matrix:
    #   - create random diagonal with elements
    #     in [0, 1]
    diag = uniform(dim)
    #   - sort it in descending order (in absolute value)
    #     [large first]
    diag = numx.take(diag, numx.argsort(abs(diag)))[::-1]
    #   - save larger elements [sfa solution]
    sfa_solution = diag[:nsources].copy()
    #   - modify diagonal elements order to allow for a
    #     different solution for isfa:
    #     create index array
    idx = list(range(0,dim))
    #     take the second slowest element and put it at the end
    idx = [idx[0]]+idx[2:]+[idx[1]]
    diag = numx.take(diag, idx)
    #   - save isfa solution
    isfa_solution = diag[:nsources]
    #   - set the first matrix
    matrices[0] = matrices[0]*diag
    # build other matrices
    diag_dim = nsources+ica_ambiguity
    for i in range(1,nmat):
        # get a random symmetric matrix
        matrices[i] = mdp.utils.symrand(dim)
        # diagonalize the subspace diag_dim
        tmp_diag = (uniform(diag_dim)-0.5)*2
        matrices[i][:diag_dim,:diag_dim] = numx.diag(tmp_diag)
    # put everything in MultCovMat
    matrices = mdp.utils.MultipleCovarianceMatrices(matrices)
    return matrices, sfa_solution, isfa_solution

def _ISFA_unmixing_error( nsources, goal, estimate):
    check = mult(goal[:nsources,:], estimate[:,:nsources])
    error = (abs(numx.sum(numx.sum(abs(check),axis=1)-1))+
             abs(numx.sum(numx.sum(abs(check),axis=0)-1)))
    error /= nsources*nsources
    return error

def testISFANode_AnalyticalSolution():
    nsources = 2
    # number of time lags
    nmat = 20
    # degree of polynomial expansion
    deg = 3
    # sfa_ica coefficient
    sfa_ica_coeff = [1., 1.]
    # how many independent subspaces in addition to the sources
    ica_ambiguity = 2
    # dimensions of expanded space
    dim = mdp.nodes._expanded_dim(deg, nsources)
    assert (nsources+ica_ambiguity) < dim, 'Too much ica ambiguity.'
    trials = 20
    for trial in range(trials):
        # get analytical solution:
        # prepared matrices, solution for sfa, solution for isf
        covs,sfa_solution,isfa_solution=_ISFA_analytical_solution(
            nsources,nmat,dim,ica_ambiguity)
        # get contrast of analytical solution
        # sfasrc, icasrc = _get_matrices_contrast(covs, nsources, dim,
        #                                         sfa_ica_coeff)
        # set rotation matrix
        R = mdp.utils.random_rot(dim)
        covs_rot = covs.copy()
        # rotate the analytical solution
        covs_rot.transform(R)
        # find the SFA solution to initialize ISFA
        eigval, SFARP = mdp.utils.symeig(covs_rot.covs[:,:,0])
        # order SFA solution by slowness
        SFARP = SFARP[:,-1::-1]
        # run ISFA
        isfa = mdp.nodes.ISFANode(lags = covs_rot.ncovs, whitened=True,
                                  sfa_ica_coeff = sfa_ica_coeff,
                                  eps_contrast = 1e-7,
                                  output_dim = nsources,
                                  max_iter = 500,
                                  verbose = False,
                                  RP = SFARP)
        isfa.train(uniform((100,dim)))
        isfa.stop_training(covs = covs_rot.copy())
        # check that the rotation matrix found by ISFA is R
        # up to a permutation matrix.
        # Unmixing error as in Tobias paper
        error = _ISFA_unmixing_error(nsources, R, isfa.RPC)
        if error < 1E-4:
            break
    assert error < 1E-4, 'None out of the %d trials succeded.' % trials



from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import str
from builtins import object
from future.utils import with_metaclass

__docformat__ = "restructuredtext en"

import pickle as _cPickle
import warnings as _warnings
import copy as _copy
import inspect

import mdp
from mdp import numx

class NodeException(mdp.MDPException):
    """Base class for exceptions in `Node` subclasses."""
    pass

class InconsistentDimException(NodeException):
    """Raised when there is a conflict setting the dimensionalities.

    Note that incoming data with conflicting dimensionality raises a normal
    `NodeException`.
    """
    pass

class TrainingException(NodeException):
    """Base class for exceptions in the training phase."""
    pass

class TrainingFinishedException(TrainingException):
    """Raised when the `Node.train` method is called although the
    training phase is closed."""
    pass

class IsNotTrainableException(TrainingException):
    """Raised when the `Node.train` method is called although the
    node is not trainable."""
    pass

class IsNotInvertibleException(NodeException):
    """Raised when the `Node.inverse` method is called although the
    node is not invertible."""
    pass


class NodeMetaclass(type):
    """A metaclass which copies docstrings from private to public methods.

    This metaclass is meant to overwrite doc-strings of methods like
    `Node.execute`, `Node.stop_training`, `Node.inverse` with the ones
    defined in the corresponding private methods `Node._execute`,
    `Node._stop_training`, `Node._inverse`, etc.

    This makes it possible for subclasses of `Node` to document the usage
    of public methods, without the need to overwrite the ancestor's methods.
    """

    # methods that can overwrite docs:
    DOC_METHODS = ['_train', '_stop_training', '_execute', '_inverse',
                   '_label', '_prob']

    def __new__(cls, classname, bases, members):
        new_cls = super(NodeMetaclass, cls).__new__(cls, classname,
                                                    bases, members)

        priv_infos = cls._select_private_methods_to_wrap(cls, members)

        # now add the wrappers
        for wrapper_name, priv_info in priv_infos.items():
            # Note: super works because we never wrap in the defining class
            orig_pubmethod = getattr(super(new_cls, new_cls), wrapper_name)

            priv_info['name'] = wrapper_name
            # preserve the last non-empty docstring
            if not priv_info['doc']:
                priv_info['doc'] = orig_pubmethod.__doc__

            recursed = hasattr(orig_pubmethod, '_undecorated_')
            if recursed:
                undec_pubmethod = orig_pubmethod._undecorated_
                priv_info.update(NodeMetaclass._get_infos(undec_pubmethod))
                wrapper_method = cls._wrap_function(undec_pubmethod,
                                                    priv_info)
                wrapper_method._undecorated_ = undec_pubmethod
            else:
                priv_info.update(NodeMetaclass._get_infos(orig_pubmethod))
                wrapper_method = cls._wrap_method(priv_info, new_cls)
                wrapper_method._undecorated_ = orig_pubmethod

            setattr(new_cls, wrapper_name, wrapper_method)
        return new_cls

    @staticmethod
    def _get_infos(pubmethod):
        infos = {}
        wrapped_info = NodeMetaclass._function_infodict(pubmethod)
        # Preserve the signature if it still does not end with kwargs
        # (this is important for binodes).
        if wrapped_info['kwargs_name'] is None:
            infos['signature'] = wrapped_info['signature']
            infos['argnames'] = wrapped_info['argnames']
            infos['defaults'] = wrapped_info['defaults']
        return infos

    @staticmethod
    def _select_private_methods_to_wrap(cls, members):
        """Select private methods that can overwrite the public docstring.

        Return a dictionary priv_infos[pubname], where the keys are the
        public name of the private method to be wrapped,
        and the values are dictionaries with the signature, doc,
        ... informations of the private methods (see `_function_infodict`).
        """
        priv_infos = {}
        for privname in cls.DOC_METHODS:
            if privname in members:
                # get the name of the corresponding public method
                pubname = privname[1:]
                # If the public method has been overwritten in this
                # subclass, then keep it.
                # This is also important because we use super in the wrapper
                # (so the public method in this class would be missed).
                if pubname not in members:
                    priv_infos[pubname] = cls._function_infodict(members[privname])
        return priv_infos

    # The next two functions (originally called get_info, wrapper)
    # are adapted versions of functions in the
    # decorator module by Michele Simionato
    # Version: 2.3.1 (25 July 2008)
    # Download page: http://pypi.python.org/pypi/decorator
    # Note: Moving these functions to utils would cause circular import.

    @staticmethod
    def _function_infodict(func):
        """
        Returns an info dictionary containing:

        - name (the name of the function : str)
        - argnames (the names of the arguments : list)
        - defaults (the values of the default arguments : tuple)
        - signature (the signature without the defaults : str)
        - doc (the docstring : str)
        - module (the module name : str)
        - dict (the function __dict__ : str)
        - kwargs_name (the name of the kwargs argument, if present, else None)

        >>> def f(self, x=1, y=2, *args, **kw): pass
        >>> info = getinfo(f)
        >>> info["name"]
        'f'
        >>> info["argnames"]
        ['self', 'x', 'y', 'args', 'kw']
        >>> info["defaults"]
        (1, 2)
        >>> info["signature"]
        'self, x, y, *args, **kw'
        >>> info["kwargs_name"]
        kw
        """
        regargs, varargs, varkwargs, defaults = inspect.getargspec(func)
        argnames = list(regargs)
        if varargs:
            argnames.append(varargs)
        if varkwargs:
            argnames.append(varkwargs)
        signature = inspect.formatargspec(regargs,
                                          varargs,
                                          varkwargs,
                                          defaults,
                                          formatvalue=lambda value: "")[1:-1]
        return dict(name=func.__name__,
                    signature=signature,
                    argnames=argnames,
                    kwargs_name=varkwargs,
                    defaults=func.__defaults__,
                    doc=func.__doc__,
                    module=func.__module__,
                    dict=func.__dict__,
                    globals=func.__globals__,
                    closure=func.__closure__)

    @staticmethod
    def _wrap_function(original_func, wrapper_infodict):
        """Return a wrapped version of func.

        :param original_func: The function to be wrapped.
        :param wrapper_infodict: The infodict to use for constructing the
            wrapper.
        """
        src = ("lambda %(signature)s: _original_func_(%(signature)s)" %
               wrapper_infodict)
        wrapped_func = eval(src, dict(_original_func_=original_func))
        wrapped_func.__name__ = wrapper_infodict['name']
        wrapped_func.__doc__ = wrapper_infodict['doc']
        wrapped_func.__module__ = wrapper_infodict['module']
        wrapped_func.__dict__.update(wrapper_infodict['dict'])
        wrapped_func.__defaults__ = wrapper_infodict['defaults']
        return wrapped_func

    @staticmethod
    def _wrap_method(wrapper_infodict, cls):
        """Return a wrapped version of func.

        :param wrapper_infodict: The infodict to be used for constructing the
            wrapper.
        :param cls: Class to which the wrapper method will be added, this is
            used for the super call.
        """
        src = ("lambda %(signature)s: super(_wrapper_class_, _wrapper_class_)."
               "%(name)s(%(signature)s)" % wrapper_infodict)
        wrapped_func = eval(src, {"_wrapper_class_": cls})
        wrapped_func.__name__ = wrapper_infodict['name']
        wrapped_func.__doc__ = wrapper_infodict['doc']
        wrapped_func.__module__ = wrapper_infodict['module']
        wrapped_func.__dict__.update(wrapper_infodict['dict'])
        wrapped_func.__defaults__ = wrapper_infodict['defaults']
        return wrapped_func


class Node(with_metaclass(NodeMetaclass, object)):
    """A `Node` is the basic building block of an MDP application.

    It represents a data processing element, like for example a learning
    algorithm, a data filter, or a visualization step.
    Each node can have one or more training phases, during which the
    internal structures are learned from training data (e.g. the weights
    of a neural network are adapted or the covariance matrix is estimated)
    and an execution phase, where new data can be processed forwards (by
    processing the data through the node) or backwards (by applying the
    inverse of the transformation computed by the node if defined).

    Nodes have been designed to be applied to arbitrarily long sets of data:
    if the underlying algorithms supports it, the internal structures can
    be updated incrementally by sending multiple batches of data (this is
    equivalent to online learning if the chunks consists of single
    observations, or to batch learning if the whole data is sent in a
    single chunk). It is thus possible to perform computations on amounts
    of data that would not fit into memory or to generate data on-the-fly.

    A `Node` also defines some utility methods, like for example
    `copy` and `save`, that return an exact copy of a node and save it
    in a file, respectively. Additional methods may be present, depending
    on the algorithm.

    `Node` subclasses should take care of overwriting (if necessary)
    the functions `is_trainable`, `_train`, `_stop_training`, `_execute`,
    `is_invertible`, `_inverse`, `_get_train_seq`, and `_get_supported_dtypes`.
    If you need to overwrite the getters and setters of the
    node's properties refer to the docstring of `get_input_dim`/`set_input_dim`,
    `get_output_dim`/`set_output_dim`, and `get_dtype`/`set_dtype`.
    """

    def __init__(self, input_dim=None, output_dim=None, dtype=None):
        """If the input dimension and the output dimension are
        unspecified, they will be set when the `train` or `execute`
        method is called for the first time.
        If dtype is unspecified, it will be inherited from the data
        it receives at the first call of `train` or `execute`.

        Every subclass must take care of up- or down-casting the internal
        structures to match this argument (use `_refcast` private
        method when possible).
        """
        # initialize basic attributes
        self._input_dim = None
        self._output_dim = None
        self._dtype = None
        # call set functions for properties
        self.set_input_dim(input_dim)
        self.set_output_dim(output_dim)
        self.set_dtype(dtype)

        # skip the training phase if the node is not trainable
        if not self.is_trainable():
            self._training = False
            self._train_phase = -1
            self._train_phase_started = False
        else:
            # this var stores at which point in the training sequence we are
            self._train_phase = 0
            # this var is False if the training of the current phase hasn't
            #  started yet, True otherwise
            self._train_phase_started = False
            # this var is False if the complete training is finished
            self._training = True

    ### properties

    def get_input_dim(self):
        """Return input dimensions."""
        return self._input_dim

    def set_input_dim(self, n):
        """Set input dimensions.

        Perform sanity checks and then calls ``self._set_input_dim(n)``, which
        is responsible for setting the internal attribute ``self._input_dim``.
        Note that subclasses should overwrite `self._set_input_dim`
        when needed.
        """
        if n is None:
            pass
        elif (self._input_dim is not None) and (self._input_dim != n):
            msg = ("Input dim are set already (%d) "
                   "(%d given)!" % (self.input_dim, n))
            raise InconsistentDimException(msg)
        else:
            self._set_input_dim(n)

    def _set_input_dim(self, n):
        self._input_dim = n

    input_dim = property(get_input_dim,
                         set_input_dim,
                         doc="Input dimensions")

    def get_output_dim(self):
        """Return output dimensions."""
        return self._output_dim

    def set_output_dim(self, n):
        """Set output dimensions.

        Perform sanity checks and then calls ``self._set_output_dim(n)``, which
        is responsible for setting the internal attribute ``self._output_dim``.
        Note that subclasses should overwrite `self._set_output_dim`
        when needed.
        """
        if n is None:
            pass
        elif (self._output_dim is not None) and (self._output_dim != n):
            msg = ("Output dim are set already (%d) "
                   "(%d given)!" % (self.output_dim, n))
            raise InconsistentDimException(msg)
        else:
            self._set_output_dim(n)

    def _set_output_dim(self, n):
        self._output_dim = n

    output_dim = property(get_output_dim,
                          set_output_dim,
                          doc="Output dimensions")

    def get_dtype(self):
        """Return dtype."""
        return self._dtype

    def set_dtype(self, t):
        """Set internal structures' dtype.

        Perform sanity checks and then calls ``self._set_dtype(n)``, which
        is responsible for setting the internal attribute ``self._dtype``.
        Note that subclasses should overwrite `self._set_dtype`
        when needed.
        """
        if t is None:
            return
        t = numx.dtype(t)
        if (self._dtype is not None) and (self._dtype != t):
            errstr = ("dtype is already set to '%s' "
                      "('%s' given)!" % (t, self.dtype.name))
            raise NodeException(errstr)
        elif t not in self.get_supported_dtypes():
            errstr = ("\ndtype '%s' is not supported.\n"
                      "Supported dtypes: %s" % (t.name,
                                                 [numx.dtype(t).name for t in
                                                  self.get_supported_dtypes()]))
            raise NodeException(errstr)
        else:
            self._set_dtype(t)

    def _set_dtype(self, t):
        t = numx.dtype(t)
        if t not in self.get_supported_dtypes():
            raise NodeException('dtype %s not among supported dtypes (%s)'
                                % (str(t), self.get_supported_dtypes()))
        self._dtype = t

    dtype = property(get_dtype,
                     set_dtype,
                     doc="dtype")

    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node.

        The types can be specified in any format allowed by :numpy:`dtype`.
        """
        # TODO: http://epydoc.sourceforge.net/manual-othermarkup.html#external-api-links for numpy
        return mdp.utils.get_dtypes('Float')

    def get_supported_dtypes(self):
        """Return dtypes supported by the node as a list of :numpy:`dtype`
        objects.

        Note that subclasses should overwrite `self._get_supported_dtypes`
        when needed."""
        return [numx.dtype(t) for t in self._get_supported_dtypes()]

    supported_dtypes = property(get_supported_dtypes,
                                doc="Supported dtypes")

    _train_seq = property(lambda self: self._get_train_seq(),
                          doc="""\
        List of tuples::

          [(training-phase1, stop-training-phase1),
           (training-phase2, stop_training-phase2),
           ...]

        By default::

          _train_seq = [(self._train, self._stop_training)]
        """)

    def _get_train_seq(self):
        return [(self._train, self._stop_training)]

    def has_multiple_training_phases(self):
        """Return True if the node has multiple training phases."""
        return len(self._train_seq) > 1

    ### Node states
    def is_training(self):
        """Return True if the node is in the training phase,
        False otherwise."""
        return self._training

    def get_current_train_phase(self):
        """Return the index of the current training phase.

        The training phases are defined in the list `self._train_seq`."""
        return self._train_phase

    def get_remaining_train_phase(self):
        """Return the number of training phases still to accomplish.

        If the node is not trainable then return 0.
        """
        if self.is_trainable():
            return len(self._train_seq) - self._train_phase
        else:
            return 0

    ### Node capabilities
    @staticmethod
    def is_trainable():
        """Return True if the node can be trained, False otherwise."""
        return True

    @staticmethod
    def is_invertible():
        """Return True if the node can be inverted, False otherwise."""
        return True

    ### check functions
    def _check_input(self, x):
        # check input rank
        if not x.ndim == 2:
            error_str = "x has rank %d, should be 2" % (x.ndim)
            raise NodeException(error_str)

        # set the input dimension if necessary
        if self.input_dim is None:
            self.input_dim = x.shape[1]

        # set the dtype if necessary
        if self.dtype is None:
            self.dtype = x.dtype

        # check the input dimension
        if not x.shape[1] == self.input_dim:
            error_str = "x has dimension %d, should be %d" % (x.shape[1],
                                                              self.input_dim)
            raise NodeException(error_str)

        if x.shape[0] == 0:
            error_str = "x must have at least one observation (zero given)"
            raise NodeException(error_str)

    def _check_output(self, y):
        # check output rank
        if not y.ndim == 2:
            error_str = "y has rank %d, should be 2" % (y.ndim)
            raise NodeException(error_str)

        # check the output dimension
        if not y.shape[1] == self.output_dim:
            error_str = "y has dimension %d, should be %d" % (y.shape[1],
                                                              self.output_dim)
            raise NodeException(error_str)

    def _if_training_stop_training(self):
        if self.is_training():
            self.stop_training()
            # if there is some training phases left we shouldn't be here!
            if self.get_remaining_train_phase() > 0:
                error_str = "The training phases are not completed yet."
                raise TrainingException(error_str)

    def _pre_execution_checks(self, x):
        """This method contains all pre-execution checks.

        It can be used when a subclass defines multiple execution methods.
        """
        # if training has not started yet, assume we want to train the node
        if (self.get_current_train_phase() == 0 and
            not self._train_phase_started):
            while True:
                self.train(x)
                if self.get_remaining_train_phase() > 1:
                    self.stop_training()
                else:
                    break

        self._if_training_stop_training()

        # control the dimension x
        self._check_input(x)

        # set the output dimension if necessary
        if self.output_dim is None:
            self.output_dim = self.input_dim

    def _pre_inversion_checks(self, y):
        """This method contains all pre-inversion checks.

        It can be used when a subclass defines multiple inversion methods.
        """
        if not self.is_invertible():
            raise IsNotInvertibleException("This node is not invertible.")

        self._if_training_stop_training()

        # set the output dimension if necessary
        if self.output_dim is None:
            # if the input_dim is not defined, raise an exception
            if self.input_dim is None:
                errstr = ("Number of input dimensions undefined. Inversion"
                          "not possible.")
                raise NodeException(errstr)
            self.output_dim = self.input_dim

        # control the dimension of y
        self._check_output(y)

    ### casting helper functions

    def _refcast(self, x):
        """Helper function to cast arrays to the internal dtype."""
        return mdp.utils.refcast(x, self.dtype)

    ### Methods to be implemented by the user

    # this are the methods the user has to overwrite
    # they receive the data already casted to the correct type

    def _train(self, x):
        if self.is_trainable():
            raise NotImplementedError

    def _stop_training(self, *args, **kwargs):
        pass

    def _execute(self, x):
        return x

    def _inverse(self, x):
        if self.is_invertible():
            return x

    def _check_train_args(self, x, *args, **kwargs):
        # implemented by subclasses if needed
        pass

    ### User interface to the overwritten methods

    def train(self, x, *args, **kwargs):
        """Update the internal structures according to the input data `x`.

        `x` is a matrix having different variables on different columns
        and observations on the rows.

        By default, subclasses should overwrite `_train` to implement their
        training phase. The docstring of the `_train` method overwrites this
        docstring.

        Note: a subclass supporting multiple training phases should implement
        the *same* signature for all the training phases and document the
        meaning of the arguments in the `_train` method doc-string. Having
        consistent signatures is a requirement to use the node in a flow.
        """

        if not self.is_trainable():
            raise IsNotTrainableException("This node is not trainable.")

        if not self.is_training():
            err_str = "The training phase has already finished."
            raise TrainingFinishedException(err_str)

        self._check_input(x)
        self._check_train_args(x, *args, **kwargs)

        self._train_phase_started = True
        self._train_seq[self._train_phase][0](self._refcast(x), *args, **kwargs)

    def stop_training(self, *args, **kwargs):
        """Stop the training phase.

        By default, subclasses should overwrite `_stop_training` to implement
        this functionality. The docstring of the `_stop_training` method
        overwrites this docstring.
        """
        if self.is_training() and self._train_phase_started == False:
            raise TrainingException("The node has not been trained.")

        if not self.is_training():
            err_str = "The training phase has already finished."
            raise TrainingFinishedException(err_str)

        # close the current phase.
        self._train_seq[self._train_phase][1](*args, **kwargs)
        self._train_phase += 1
        self._train_phase_started = False
        # check if we have some training phase left
        if self.get_remaining_train_phase() == 0:
            self._training = False

    def execute(self, x, *args, **kwargs):
        """Process the data contained in `x`.

        If the object is still in the training phase, the function
        `stop_training` will be called.
        `x` is a matrix having different variables on different columns
        and observations on the rows.

        By default, subclasses should overwrite `_execute` to implement
        their execution phase. The docstring of the `_execute` method
        overwrites this docstring.
        """
        self._pre_execution_checks(x)
        return self._execute(self._refcast(x), *args, **kwargs)

    def inverse(self, y, *args, **kwargs):
        """Invert `y`.

        If the node is invertible, compute the input ``x`` such that
        ``y = execute(x)``.

        By default, subclasses should overwrite `_inverse` to implement
        their `inverse` function. The docstring of the `inverse` method
        overwrites this docstring.
        """
        self._pre_inversion_checks(y)
        return self._inverse(self._refcast(y), *args, **kwargs)

    def __call__(self, x, *args, **kwargs):
        """Calling an instance of `Node` is equivalent to calling
        its `execute` method."""
        return self.execute(x, *args, **kwargs)

    ###### adding nodes returns flows

    def __add__(self, other):
        # check other is a node
        if isinstance(other, Node):
            return mdp.Flow([self, other])
        elif isinstance(other, mdp.Flow):
            flow_copy = other.copy()
            flow_copy.insert(0, self)
            return flow_copy.copy()
        else:
            err_str = ('can only concatenate node'
                       ' (not \'%s\') to node' % (type(other).__name__))
            raise TypeError(err_str)

    ###### string representation

    def __str__(self):
        return str(type(self).__name__)

    def __repr__(self):
        # print input_dim, output_dim, dtype
        name = type(self).__name__
        inp = "input_dim=%s" % str(self.input_dim)
        out = "output_dim=%s" % str(self.output_dim)
        if self.dtype is None:
            typ = 'dtype=None'
        else:
            typ = "dtype='%s'" % self.dtype.name
        args = ', '.join((inp, out, typ))
        return name + '(' + args + ')'

    def copy(self, protocol=None):
        """Return a deep copy of the node.

        :param protocol: the pickle protocol (deprecated)."""
        if protocol is not None:
            _warnings.warn("protocol parameter to copy() is ignored",
                           mdp.MDPDeprecationWarning, stacklevel=2)
        return _copy.deepcopy(self)

    def save(self, filename, protocol=-1):
        """Save a pickled serialization of the node to `filename`.
        If `filename` is None, return a string.

        Note: the pickled `Node` is not guaranteed to be forwards or
        backwards compatible."""
        if filename is None:
            return _cPickle.dumps(self, protocol)
        else:
            # if protocol != 0 open the file in binary mode
            mode = 'wb' if protocol != 0 else 'w'
            with open(filename, mode) as flh:
                _cPickle.dump(self, flh, protocol)


class PreserveDimNode(Node):
    """Abstract base class with ``output_dim == input_dim``.

    If one dimension is set then the other is set to the same value.
    If the dimensions are set to different values, then an
    `InconsistentDimException` is raised.
    """

    def _set_input_dim(self, n):
        if (self._output_dim is not None) and (self._output_dim != n):
            err = "input_dim must be equal to output_dim for this node."
            raise InconsistentDimException(err)
        self._input_dim = n
        self._output_dim = n

    def _set_output_dim(self, n):
        if (self._input_dim is not None) and (self._input_dim != n):
            err = "output_dim must be equal to input_dim for this node."
            raise InconsistentDimException(err)
        self._input_dim = n
        self._output_dim = n


def VariadicCumulator(*fields):
    """A VariadicCumulator is a `Node` whose training phase simply collects
    all input data. In this way it is possible to easily implement
    batch-mode learning.

    The data is accessible in the attributes given with the VariadicCumulator's
    constructor after the beginning of the `Node._stop_training` phase.
    ``self.tlen`` contains the number of data points collected.
    """

    class Cumulator(Node):
        def __init__(self, *args, **kwargs):
            super(Cumulator, self).__init__(*args, **kwargs)
            self._cumulator_fields = fields
            for arg in self._cumulator_fields:
                if hasattr(self, arg):
                    errstr = "Cumulator Error: Property %s already defined"
                    raise mdp.MDPException(errstr % arg)
                setattr(self, arg, [])
            self.tlen = 0

        def _train(self, *args):
            """Collect all input data in a list."""
            self.tlen += args[0].shape[0]
            for field, data in zip(self._cumulator_fields, args):
                getattr(self, field).append(data)

        def _stop_training(self, *args, **kwargs):
            """Concatenate the collected data in a single array."""
            for field in self._cumulator_fields:
                data = getattr(self, field)
                setattr(self, field, numx.concatenate(data, 0))

    return Cumulator

Cumulator = VariadicCumulator('data')
Cumulator.__doc__ = """A specialized version of `VariadicCumulator` which only
                    fills the field ``self.data``.
                    """



from __future__ import division
from builtins import range
from past.utils import old_div
from ._tools import *

def test_basic_training():
    dim = 10000
    freqs = [2*numx.pi*100.,2*numx.pi*500.]
    t =  numx.linspace(0,1,num=dim)
    mat = numx.array([numx.sin(freqs[0]*t),numx.sin(freqs[1]*t)]).T
    mat += normal(0., 1e-10, size=(dim, 2))
    mat = old_div((mat - mean(mat[:-1,:],axis=0)),std(mat[:-1,:],axis=0))
    des_mat = mat.copy()
    mat = mult(mat,uniform((2,2))) + uniform(2)
    sfa = mdp.nodes.SFA2Node()
    sfa.train(mat)
    out = sfa.execute(mat)
    assert out.shape[1]==5, "Wrong output_dim"
    correlation = old_div(mult(des_mat[:-1,:].T,
                       numx.take(out[:-1,:], (0,2), axis=1)),(dim-2))
    assert_array_almost_equal(abs(correlation),
                              numx.eye(2), 3)
    for nr in range(sfa.output_dim):
        qform = sfa.get_quadratic_form(nr)
        outq = qform.apply(mat)
        assert_array_almost_equal(outq, out[:,nr], decimal-1)

    sfa = mdp.nodes.SFANode(output_dim = 2)
    sfa.train(mat)
    out = sfa.execute(mat)
    assert out.shape[1]==2, 'Wrong output_dim'
    correlation = old_div(mult(des_mat[:-1,:1].T,out[:-1,:1]),(dim-2))
    assert_array_almost_equal(abs(correlation),
                              numx.eye(1), 3)

def test_range_argument():
    node = mdp.nodes.SFA2Node()
    x = numx.random.random((100,10))
    node.train(x)
    node.stop_training()
    y = node.execute(x, n=5)
    assert y.shape[1] == 5

def test_input_dim_bug():
    dim = 10000
    freqs = [2*numx.pi*100.,2*numx.pi*500.]
    t =  numx.linspace(0,1,num=dim)
    mat = numx.array([numx.sin(freqs[0]*t),numx.sin(freqs[1]*t)]).T
    mat += normal(0., 1e-10, size=(dim, 2))
    mat = old_div((mat - mean(mat[:-1,:],axis=0)),std(mat[:-1,:],axis=0))
    mat = mult(mat,uniform((2,2))) + uniform(2)
    sfa = mdp.nodes.SFA2Node(input_dim=2)
    sfa.train(mat)
    sfa.execute(mat)

def test_output_dim_bug():
    dim = 10000
    freqs = [2*numx.pi*100.,2*numx.pi*500.]
    t =  numx.linspace(0,1,num=dim)
    mat = numx.array([numx.sin(freqs[0]*t),numx.sin(freqs[1]*t)]).T
    mat += normal(0., 1e-10, size=(dim, 2))
    mat = old_div((mat - mean(mat[:-1,:],axis=0)), std(mat[:-1,:],axis=0))
    mat = mult(mat,uniform((2,2))) + uniform(2)
    sfa = mdp.nodes.SFA2Node(output_dim=3)
    sfa.train(mat)
    out = sfa.execute(mat)
    assert out.shape[1] == 3



__docformat__ = "restructuredtext en"

from mdp import numx, numx_linalg, utils, NodeException
import mdp
import scipy.signal as signal

# TODO automatic selection of convolution

# TODO provide generators for standard filters

# TODO look into Theano, define TheanoConvolutionNode

class Convolution2DNode(mdp.Node):
    """Convolve input data with filter banks.

    The ``filters`` argument specifies a set of 2D filters that are
    convolved with the input data during execution. Convolution can
    be selected to be executed by linear filtering of the data, or
    in the frequency domain using a Discrete Fourier Transform.

    Input data can be given as 3D data, each row being a 2D array
    to be convolved with the filters, or as 2D data, in which case
    the ``input_shape`` argument must be specified.

    This node depends on ``scipy``.
    """

    def __init__(self, filters, input_shape = None,
                 approach = 'fft',
                 mode = 'full', boundary = 'fill', fillvalue = 0,
                 output_2d = True,
                 input_dim = None, dtype = None):
        """
        Input arguments:

        input_shape -- Is a tuple (h,w) that corresponds to the height and
                       width of the input 2D data. If the input data is given
                       in a flattened format, it is first reshaped before
                       convolution

        approach -- 'approach' is one of ['linear', 'fft']
                    'linear': convolution is done by linear filtering;
                    'fft': convoltion is done using the Fourier Transform
                    If 'approach' is 'fft', the 'boundary' and 'fillvalue' arguments
                    are ignored, and are assumed to be 'fill' and 0, respectively.
                    (*Default* = 'fft')

        mode -- Convolution mode, as defined in scipy.signal.convolve2d
                'mode' is one of ['valid', 'same', 'full']
                (*Default* = 'full')

        boundary -- Boundary condition, as defined in scipy.signal.convolve2d
                     'boundary' is one of ['fill', 'wrap', 'symm']
                     (*Default* = 'fill')

        fillvalue -- Value to fill pad input arrays with
                     (*Default* = 0)

        output_2d -- If True, the output array is 2D; the first index
                     corresponds to data points; every output data point
                     is the result of flattened convolution results, with
                     the output of each filter concatenated together.

                     If False, the output array is 4D; the format is
                     data[idx,filter_nr,x,y], with
                     filter_nr: index of convolution filter
                     idx: data point index
                     x, y: 2D coordinates
        """
        super(Convolution2DNode, self).__init__(input_dim=input_dim,
                                              dtype=dtype)

        self.filters = filters

        self._input_shape = input_shape

        if approach not in ['linear', 'fft']:
            raise NodeException("'approach' argument must be one of ['linear', 'fft']")
        self._approach = approach

        if mode not in ['valid', 'same', 'full']:
            raise NodeException("'mode' argument must be one of ['valid', 'same', 'full']")
        self._mode = mode

        self.boundary = boundary
        self.fillvalue = fillvalue
        self.output_2d = output_2d
        self._output_shape = None

    # ------- class properties
    def get_filters(self):
        return self._filters

    def set_filters(self, filters):
        if not isinstance(filters, numx.ndarray):
            raise NodeException("'filters' argument must be a numpy array")
        if filters.ndim != 3:
            raise NodeException('Filters must be specified in a 3-dim array, with each '+
                                'filter on a different row')
        self._filters = filters

    filters = property(get_filters, set_filters)

    def get_boundary(self):
        return self._boundary

    def set_boundary(self, boundary):
        if boundary not in ['fill', 'wrap', 'symm']:
            raise NodeException(
                "'boundary' argument must be one of ['fill', 'wrap', 'symm']")
        self._boundary = boundary

    boundary = property(get_boundary, set_boundary)

    @property
    def input_shape(self):
        return self._input_shape

    @property
    def approach(self):
        return self._approach

    @property
    def mode(self):
        return self._mode

    @property
    def output_shape(self):
        return self._output_shape
    # ------- /class properties

    def is_trainable(self):
        return False

    def is_invertible(self):
        return False

    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node.
      
        Support floating point types with size smaller or equal than 64 bits.
        This is because fftpack does not support floating point types larger
        than that.
        """
        return [t for t in utils.get_dtypes('Float') if t.itemsize<=8]

    def _pre_execution_checks(self, x):
        """This method contains all pre-execution checks.
        It can be used when a subclass defines multiple execution methods.

        In this case, the output dimension depends on the type of
        convolution we use (padding, full, ...). Also, we want to
        to be able to accept 3D arrays.
        """

        # check input rank
        if not x.ndim in [2,3]:
            error_str = "x has rank %d, should be 2 or 3" % (x.ndim)
            raise NodeException(error_str)

        # set 2D shape if necessary
        if self._input_shape is None:
            if x.ndim == 2:
                error_str = "Cannot infer 2D shape from 1D data points. " + \
                            "Data must have rank 3, or shape argument given."
                raise NodeException(error_str)
            else:
                self._input_shape = x.shape[1:]

        # set the input dimension if necessary
        if self.input_dim is None:
            self.input_dim = numx.prod(self._input_shape)

        # set the dtype if necessary
        if self.dtype is None:
            self.dtype = x.dtype

        # check the input dimension
        if not numx.prod(x.shape[1:]) == self.input_dim:
            error_str = "x has dimension %d, should be %d" % (x.shape[1],
                                                              self.input_dim)
            raise NodeException(error_str)

        # set output_dim if necessary
        if self.output_dim is None:
            input_shape = self.input_shape
            filters_shape = self.filters.shape
            if self.mode == 'same':
                self._output_shape = input_shape
            elif self.mode == 'full':
                self._output_shape = (input_shape[0]+filters_shape[1]-1,
                                      input_shape[1]+filters_shape[2]-1)
            else: # mode == 'valid'
                self._output_shape = (input_shape[0]-filters_shape[1]+1,
                                      input_shape[1]-filters_shape[2]+1)
            self.output_dim = self.filters.shape[0]*numx.prod(self._output_shape)

        if x.shape[0] == 0:
            error_str = "x must have at least one observation (zero given)"
            raise NodeException(error_str)

    def _execute(self, x):
        is_2d = x.ndim==2
        output_shape, input_shape = self._output_shape, self._input_shape
        filters = self.filters
        nfilters = filters.shape[0]

        # XXX depends on convolution
        y = numx.empty((x.shape[0], nfilters,
                        output_shape[0], output_shape[1]), dtype=self.dtype)
        for n_im, im in enumerate(x):
            if is_2d:
                im = im.reshape(input_shape)
            for n_flt, flt in enumerate(filters):
                if self.approach == 'fft':
                    y[n_im,n_flt,:,:] = signal.fftconvolve(im, flt, mode=self.mode)
                elif self.approach == 'linear':
                    y[n_im,n_flt,:,:] = signal.convolve2d(im, flt,
                                                          mode=self.mode,
                                                          boundary=self.boundary,
                                                          fillvalue=self.fillvalue)

        # reshape if necessary
        if self.output_2d:
            y.resize((y.shape[0], self.output_dim))

        return y



from builtins import range
from builtins import object
import mdp
from mdp import numx as n

from bimdp import BiFlow, MSG_ID_SEP, EXIT_TARGET
from bimdp.hinet import BiFlowNode, CloneBiLayer, BiSwitchboard
from bimdp.nodes import SFABiNode, IdentityBiNode


class TestBiFlowNode(object):
    """Test the behavior of the BiFlowNode."""

    def test_two_nodes1(self):
        """Test a TestBiFlowNode with two normal nodes."""
        sfa_node = mdp.nodes.SFANode(input_dim=10, output_dim=8)
        sfa2_node = mdp.nodes.SFA2Node(input_dim=8, output_dim=6)
        flownode = BiFlowNode(BiFlow([sfa_node, sfa2_node]))
        for _ in range(2):
            for _ in range(6):
                flownode.train(n.random.random((30,10)))
            flownode.stop_training()
        x = n.random.random([100,10])
        flownode.execute(x)

    def test_two_nodes2(self):
        """Test a TestBiFlowNode with two normal nodes using a normal Flow."""
        sfa_node = mdp.nodes.SFANode(input_dim=10, output_dim=8)
        sfa2_node = mdp.nodes.SFA2Node(input_dim=8, output_dim=6)
        flownode = BiFlowNode(BiFlow([sfa_node, sfa2_node]))
        flow = mdp.Flow([flownode])
        data_iterables = [[n.random.random((30,10)) for _ in range(6)]]
        flow.train(data_iterables)
        x = n.random.random([100,10])
        flow.execute(x)

    def test_pretrained_nodes(self):
        """Test a TestBiFlowNode with two normal pretrained nodes."""
        sfa_node = mdp.nodes.SFANode(input_dim=10, output_dim=8)
        sfa2_node = mdp.nodes.SFA2Node(input_dim=8, output_dim=6)
        flownode = BiFlowNode(BiFlow([sfa_node, sfa2_node]))
        flow = mdp.Flow([flownode])
        data_iterables = [[n.random.random((30,10)) for _ in range(6)]]
        flow.train(data_iterables)
        pretrained_flow = flow[0]._flow
        biflownode = BiFlowNode(pretrained_flow)
        x = n.random.random([100,10])
        biflownode.execute(x)


class DummyBiNode(IdentityBiNode):
    """Dummy class for CloneBiLayer tests."""

    def _execute(self, x, data1, data2):
        self.data1 = data1
        self.data2 = data2
        return x

    @staticmethod
    def is_trainable():
        return False


class TestCloneBiLayer(object):
    """Test the behavior of the BiCloneLayer."""

    def test_clonelayer(self):
        """Test a simple clonelayer with three SFA Nodes."""
        sfa_node = SFABiNode(input_dim=3, output_dim=2)
        clonelayer = CloneBiLayer(sfa_node, 3)
        x = n.random.random((100,9))
        clonelayer.train(x)
        clonelayer.stop_training()
        clonelayer.execute(x)

    def test_use_copies_msg(self):
        """Test the correct reaction to an outgoing use_copies message."""
        stop_result = ({"clonelayer" + MSG_ID_SEP + "use_copies": True}, 1)
        stop_sfa_node = SFABiNode(stop_result=stop_result,
                                  input_dim=10, output_dim=3)
        clonelayer = CloneBiLayer(node=stop_sfa_node,
                                  n_nodes=3,
                                  use_copies=False,
                                  node_id="clonelayer")
        x = n.random.random((100,30))
        clonelayer.train(x)
        clonelayer.stop_training()
        assert clonelayer.use_copies is True

    def test_use_copies_msg_flownode(self):
        """Test the correct reaction to an outgoing use_copies message."""
        stop_result = ({"clonelayer" + MSG_ID_SEP + "use_copies": True},
                       EXIT_TARGET)
        stop_sfa_node = SFABiNode(stop_result=stop_result,
                                  input_dim=10, output_dim=3)
        biflownode = BiFlowNode(BiFlow([stop_sfa_node]))
        clonelayer = CloneBiLayer(node=biflownode,
                                  n_nodes=3,
                                  use_copies=False,
                                  node_id="clonelayer")
        biflow = clonelayer + IdentityBiNode()
        x = n.random.random((100,30))
        biflow.train(x)
        assert clonelayer.use_copies is True

    def test_message_splitting(self):
        """Test message array splitting and combination."""
        node = DummyBiNode(input_dim=3)
        clonelayer = CloneBiLayer(node, 2, use_copies=True)
        x = n.random.random((10, 6))
        data1 = n.random.random((10, 4))  # should be split
        data2 = n.random.random((10, 5))  # should not be touched
        msg = {
            "string": "blabla",
            "list": [1,2],
            "data1": data1,
            "data2": data2,
        }
        y, out_msg = clonelayer.execute(x, msg)
        node1, node2 = clonelayer.nodes
        assert n.all(x == y)
        assert out_msg["string"] == msg["string"]
        assert out_msg["list"] == msg["list"]
        assert n.all(out_msg["data1"] == data1)
        assert n.all(node1.data1 == data1[:,:2])
        assert n.all(node2.data1 == data1[:,2:])
        assert out_msg["data2"] is data2
        assert n.all(node1.data2 is data2)
        assert n.all(node2.data2 is data2)


class TestBiSwitchboardNode(object):
    """Test the behavior of the BiSwitchboardNode."""

    def test_execute_routing(self):
        """Test the standard routing for messages."""
        sboard = BiSwitchboard(input_dim=3, connections=[2,0,1])
        x = n.array([[1,2,3],[4,5,6]])
        msg = {
            "string": "blabla",
            "list": [1,2],
            "data": x.copy(),  # should be mapped by switchboard
            "data2": n.zeros(3),  # should not be modified
            "data3": n.zeros((3,4)),  # should not be modified
        }
        y, out_msg = sboard.execute(x, msg)
        reference_y = n.array([[3,1,2],[6,4,5]])
        assert (y == reference_y).all()
        assert out_msg["string"] == msg["string"]
        assert out_msg["list"] == msg["list"]
        assert n.all(out_msg["data"] == reference_y)
        assert out_msg["data2"].shape == (3,)
        assert out_msg["data3"].shape == (3,4)

    def test_inverse_message_routing(self):
        """Test the inverse routing for messages."""
        sboard = BiSwitchboard(input_dim=3, connections=[2,0,1])
        x = n.array([[1,2,3],[4,5,6]])
        msg = {
            "string": "blabla",
            "method": "inverse",
            "list": [1,2],
            "data": x,  # should be mapped by switchboard
            "data2": n.zeros(3),  # should not be modified
            "data3": n.zeros((3,4)),  # should not be modified
            "target": "test"
        }
        y, out_msg, target = sboard.execute(None, msg)
        assert y is None
        assert target == "test"
        reference_y = n.array([[2,3,1],[5,6,4]])
        assert out_msg["string"] == msg["string"]
        assert out_msg["list"] == msg["list"]
        assert (out_msg["data"] == reference_y).all()
        assert out_msg["data2"].shape == (3,)
        assert out_msg["data3"].shape == (3,4)
